import datetime
import sqlite3
from os.path import expanduser
from typing import Iterable, List

from prompt_toolkit.history import History

from peek.config import config_location, ensure_dir_exists

HIST_MAX = 10_000


class SqLiteHistory(History):
    def __init__(self, history_max=HIST_MAX):
        super().__init__()
        self.history_max = history_max
        db_file = expanduser(config_location() + 'history')
        ensure_dir_exists(db_file)
        self.conn = sqlite3.connect(db_file)
        self.conn.execute(
            'CREATE TABLE IF NOT EXISTS history '
            '(id INTEGER PRIMARY KEY AUTOINCREMENT, content TEXT NOT NULL, timestamp INTEGER NOT NULL)'
        )
        self.conn.execute(
            'CREATE TABLE IF NOT EXISTS connection '
            '(name TEXT PRIMARY KEY, data TEXT NOT NULL, timestamp INTEGER NOT NULL)'
        )
        self._maintain_size()
        self.conn.commit()

    def __del__(self):
        self.conn.close()

    def _maintain_size(self):
        c = self.conn.cursor()
        res = c.execute('SELECT COUNT(*) from history').fetchone()
        if res is None or res[0] < self.history_max:
            return
        c.execute(
            'DELETE FROM history where id in (SELECT id FROM history ORDER BY id limit ?)', (res[0] - self.history_max,)
        )

    def load_history_strings(self) -> Iterable[str]:
        strings: List[str] = []
        for row in self.conn.execute('SELECT * FROM history ORDER BY id DESC'):
            strings.append(row[1])
        return strings

    def store_string(self, string: str) -> None:
        self.conn.execute("INSERT INTO history(content, timestamp) VALUES (?, ?)", (string, datetime.datetime.now()))
        self.conn.commit()

    def load_recent(self, size=100):
        lines = []
        for row in self.conn.execute("select * from history where id > (select max(id) from history) - ?", (size,)):
            lines.append((row[0], row[1]))
        return lines

    def get_entry(self, index):
        if index > 0:
            for row in self.conn.execute('select * from history where id = ?', (index,)):
                return row[0], row[1]
            else:
                return None
        elif index < 0:
            for row in self.conn.execute(
                'select * from history where id = (select max(id) from history) + ?', (index,)
            ):
                return row[0], row[1]
            else:
                return None
        else:
            return None

    def save_session(self, name, data):
        self.conn.execute(
            "INSERT INTO connection(name, data, timestamp) VALUES(?, ?, ?) "
            "ON CONFLICT (name) DO UPDATE SET data=excluded.data, timestamp=excluded.timestamp",
            (name, data, datetime.datetime.now()),
        )
        self.conn.commit()

    def load_session(self, name):
        for row in self.conn.execute("SELECT data FROM connection WHERE name = ?", (name,)):
            return row[0]
        else:
            return None

    def list_sessions(self):
        return list(self.conn.execute("SELECT name, timestamp FROM connection"))

    def clear_sessions(self):
        self.conn.execute("DELETE FROM connection")

    def delete_session(self, name):
        return self.conn.execute("DELETE FROM connection where name = ?", (name,)).rowcount == 1



import logging
import numbers
from dataclasses import dataclass
from typing import Any, Dict, List, Union

_logger = logging.getLogger(__name__)

Types = "Dict[TypeName, Union[Builtin, Alias, Interface, Enum, Request]]"


@dataclass(frozen=True)
class TypeName:
    name: str
    namespace: str

    @staticmethod
    def from_dict(data: Dict):
        return TypeName(data['name'], data['namespace'])


@dataclass(frozen=True)
class Endpoint:
    urls: List[Dict]
    request: TypeName
    description: str
    doc_url: str

    @staticmethod
    def from_dict(data: Dict):
        return Endpoint(
            urls=data['urls'],
            request=TypeName.from_dict(data['request']) if data['request'] else None,
            description=data['description'],
            doc_url=data['docUrl'],
        )


@dataclass(frozen=True)
class TypeDefinition:
    name: TypeName
    data: Dict

    def candidate_values(self, types: Types):
        yield from []

    def candidate_properties(self, types: Types):
        return []

    def __getattr__(self, item):
        return self.data.get(item, None)

    @staticmethod
    def from_dict(data: Dict):
        name = TypeName.from_dict(data.pop('name'))
        kind = data.pop('kind')
        if kind == 'type_alias':
            return Alias(name=name, data=data)
        elif kind == 'interface':
            return Interface(name=name, data=data)
        elif kind == 'enum':
            return Enum(name=name, data=data)
        elif kind == 'request':
            return Request(name=name, data=data)
        elif kind == 'response':
            return Response(name=name, data=data)
        else:
            raise ValueError(f'unrecognized type definition kind [{kind}]')


@dataclass(frozen=True)
class Builtin(TypeDefinition):
    def candidate_values(self, types: Types):
        if self.name.name == 'boolean':
            yield from [True, False]
        elif self.name.name == 'string':
            yield from ['']
        elif self.name.name == 'null':
            yield from [None]
        elif self.name.name == 'number':
            yield from [0]
        else:
            yield from []


@dataclass(frozen=True)
class Alias(TypeDefinition):
    def candidate_values(self, types: Types):
        yield from self.get_type().candidate_values(types)

    def candidate_properties(self, types: Types):
        return self.get_type().candidate_properties(types)

    def get_type(self):
        return Value.from_dict(self.type)


@dataclass(frozen=True)
class Interface(TypeDefinition):
    def candidate_values(self, types: Types):
        yield from [{}]

    def candidate_properties(self, types: Types):
        return [Variable.from_dict(prop) for prop in self.properties]


@dataclass(frozen=True)
class Enum(TypeDefinition):
    def candidate_values(self, types: Types):
        yield from self.get_members()

    def get_members(self) -> List[str]:
        return [member['name'] for member in self.members]


@dataclass(frozen=True)
class Request(TypeDefinition):
    def get_query(self) -> Dict:
        query = self.query
        if query is not None:
            if not isinstance(query, dict):
                query = {q['name']: q for q in self.query}
                self.data['query'] = query
        return query

    def get_attached_behaviours(self) -> List[str]:
        return self.attachedBehaviors


@dataclass(frozen=True)
class Response(TypeDefinition):
    pass


@dataclass(frozen=True)
class Value:
    data: Dict

    def candidate_values(self, types: Types):
        yield from []

    def candidate_properties(self, types: Types):
        return []

    def __getattr__(self, item):
        return self.data.get(item, None)

    @staticmethod
    def from_dict(data):
        kind = data['kind']
        if kind == 'instance_of':
            return InstanceOf(data=data)
        elif kind == 'array_of':
            return ArrayOf(data=data)
        elif kind == 'dictionary_of':
            return DictionaryOf(data=data)
        elif kind == 'union_of':
            return UnionOf(data=data)
        elif kind == 'literal_value':
            return Literal(data=data)
        elif kind == 'user_defined_value':
            return UserDefined(data=data)
        elif kind == 'void_value':
            return Void(data=data)
        else:
            raise ValueError(f'unrecognized value kind [{kind}]')


@dataclass(frozen=True)
class InstanceOf(Value):
    def candidate_values(self, types: Types):
        type_name = self.get_type_name()
        if type_name in types:
            yield from types[type_name].candidate_values(types)
        else:
            _logger.debug(f'type [{type_name}] does not exist')
            yield from []

    def candidate_properties(self, types: Types):
        type_name = self.get_type_name()
        if type_name in types:
            return types[type_name].candidate_properties(types)
        else:
            _logger.debug(f'type [{type_name}] does not exist')
            return []

    def get_type_name(self) -> TypeName:
        return TypeName.from_dict(self.type)


@dataclass(frozen=True)
class ArrayOf(Value):
    def candidate_values(self, types: Types):
        member_value = next(self.get_member().candidate_values(types))
        if member_value == {}:
            yield from [[{}]]
        else:
            yield from [[]]

    def get_member(self) -> Value:
        return Value.from_dict(self.value)


@dataclass(frozen=True)
class DictionaryOf(Value):
    def candidate_values(self, types: Types):
        yield from [{}]

    def candidate_properties(self, types: Types):
        # TODO: check key type is string?
        return [Wildcard(self.get_value())]

    def get_key(self) -> Value:
        return Value.from_dict(self.key)

    def get_value(self) -> Value:
        return Value.from_dict(self.value)

    def is_single_key(self) -> bool:
        return self.singleKey


@dataclass(frozen=True)
class UnionOf(Value):
    def candidate_values(self, types: Types):
        for member in self.get_members():
            yield from member.candidate_values(types)

    def candidate_properties(self, types: Types):
        all_results = []
        for member in self.get_members():
            all_results.extend(member.candidate_properties(types))
        return all_results

    def get_members(self):
        return [Value.from_dict(item) for item in self.items]


@dataclass(frozen=True)
class Literal(Value):
    def candidate_values(self, types: Types):
        yield from [self.get_value()]

    def get_value(self):
        return self.value


@dataclass(frozen=True)
class UserDefined(Value):
    pass


@dataclass(frozen=True)
class Void(Value):
    pass


@dataclass(frozen=True)
class Variable:
    name: str
    aliases: List[str]
    value: Value

    def candidate_values(self, types: Types):
        yield from self.value.candidate_values(types)

    def candidate_properties(self, types: Types):
        return self.value.candidate_properties(types)

    def match_key(self, key: str):
        return self.name == key or key in self.aliases

    @staticmethod
    def from_dict(data):
        return Variable(name=data['name'], aliases=data.get('aliases', []), value=Value.from_dict(data['type']))


@dataclass(frozen=True, init=False)
class Wildcard(Variable):
    def __init__(self, value: Value):
        super(Wildcard, self).__init__(name='*', aliases=[], value=value)

    def match_key(self, key: str):
        return True


@dataclass(frozen=True)
class Body:
    data: Dict

    def candidate_properties(self) -> List[Variable]:
        return []

    def __getattr__(self, item):
        return self.data.get(item, None)

    @staticmethod
    def from_dict(data):
        kind = data['kind']
        if kind == 'no_body':
            return NoBody(data)
        elif kind == 'value':
            return ValueBody(data)
        elif kind == 'properties':
            return PropertiesBody(data)
        else:
            raise ValueError(f'unrecognized body kind [{kind}]')


@dataclass(frozen=True)
class NoBody(Body):
    pass


@dataclass(frozen=True)
class ValueBody(Body):
    def get_value(self):
        return Value.of(self.value)


@dataclass(frozen=True)
class PropertiesBody(Body):
    def candidate_properties(self) -> List[Variable]:
        return [Variable.from_dict(prop) for prop in self.properties]


class Schema:
    def __init__(self, data: Dict):
        self.endpoints = [Endpoint.from_dict(d) for d in data['endpoints']]
        self.types: Types = {
            b.name: b
            for b in [
                Builtin(name=TypeName("binary", "_builtins"), data={}),
                Builtin(name=TypeName("boolean", "_builtins"), data={}),
                Builtin(name=TypeName("null", "_builtins"), data={}),
                Builtin(name=TypeName("number", "_builtins"), data={}),
                Builtin(name=TypeName("string", "_builtins"), data={}),
                Builtin(name=TypeName("void", "_builtins"), data={}),
            ]
        }
        for d in data['types']:
            type_definition = TypeDefinition.from_dict(d)
            if isinstance(type_definition, Response):
                continue
            self.types[type_definition.name] = type_definition
        self._common_parameters = self._build_common_params()

    def candidate_urls(self, method: str, ts: List[str]) -> List[str]:
        candidates = []
        for endpoint in self.endpoints:
            for url in endpoint.urls:
                if method not in url['methods']:
                    continue
                ps = [p for p in url['path'].split('/') if p]
                # Nothing to complete if the candidate is shorter than current input
                if len(ts) >= len(ps):
                    continue
                if not self._can_match(ts, ps):
                    continue
                candidate = '/'.join(ps[len(ts) :])
                candidates.append(candidate)
        return sorted(candidates)

    def candidate_query_param_names(self, method: str, ts: List[str]) -> List[str]:
        candidates = set()
        for endpoint in self._matchable_endpoints(method, ts):
            if endpoint.request is None:
                continue
            request: Request = self.types[endpoint.request]
            if self._request_has_common_query_params(request):
                candidates.update(self._common_parameters.keys())
            query = request.get_query()
            if query is not None:
                candidates.update(query.keys())

        return sorted(candidates)

    def candidate_query_param_values(self, method: str, ts: List[str], param_name: str) -> List[str]:
        candidates = set()
        for endpoint in self._matchable_endpoints(method, ts):
            if endpoint.request is None:
                continue
            request: Request = self.types[endpoint.request]
            if self._request_has_common_query_params(request) and param_name in self._common_parameters.keys():
                candidates.update(self._common_parameters[param_name])
            else:
                query_param = self._get_query_parameter(param_name, request)
                if query_param is not None:
                    candidates.update(self._filter_for_param_values(query_param.candidate_values(self.types)))

        return sorted(candidates)

    def candidate_sub_key_values(self, method, ts: List[str], payload_keys: List[str]) -> Dict[str, Any]:
        endpoint: Endpoint = self._matchable_endpoint(method, ts)
        if endpoint is None or endpoint.request is None:
            return {}
        sub_properties = self._sub_properties_for_keys(Body.from_dict(self.types[endpoint.request].body), payload_keys)
        key_to_values = {}
        for sub_prop in sub_properties:
            # Filter out wildcard to avoid showing '*' as a suggestion for dict keys
            if not isinstance(sub_prop, Variable) or isinstance(sub_prop, Wildcard):
                continue
            try:
                first_candidate_value = next(sub_prop.candidate_values(self.types))
            except StopIteration:
                # Some properties do not have candidate values, e.g. binary property
                # Just catch them all with an empty string
                first_candidate_value = ''

            key_to_values[sub_prop.name] = first_candidate_value
            for alias in sub_prop.aliases:
                key_to_values[alias] = first_candidate_value

        return key_to_values

    def candidate_values(self, method, ts: List[str], payload_keys: List[str], inside_array=False) -> List[Any]:
        endpoint: Endpoint = self._matchable_endpoint(method, ts)
        if endpoint is None or endpoint.request is None:
            return []
        properties = self._properties_for_keys(Body.from_dict(self.types[endpoint.request].body), payload_keys)
        values = []
        for prop in properties:
            if inside_array and isinstance(prop.value, ArrayOf):
                # penetrate array
                values.extend(prop.value.get_member().candidate_values(self.types))
            else:
                values.extend(prop.candidate_values(self.types))
        return values

    def _matchable_endpoints(self, method: str, ts: List[str]):
        for endpoint in self.endpoints:
            matched = False
            for url in endpoint.urls:
                if method not in url['methods']:
                    continue
                ps = [p for p in url['path'].split('/') if p]
                if len(ts) != len(ps):
                    continue
                if not self._can_match(ts, ps):
                    continue
                matched = True
                break
            if matched:
                yield endpoint

    def _matchable_endpoint(self, method, ts: List[str]) -> Union[Endpoint, None]:
        try:
            endpoint = next(self._matchable_endpoints(method, ts))
            _logger.debug(f'Found endpoint for {method!r} {ts}')
            return endpoint
        except StopIteration:
            _logger.debug(f'No matching endpoint found for {method!r} {ts}')
            return None

    def _properties_for_keys(self, body: Body, payload_keys: List[str]) -> List[Variable]:
        """
        For the given payload_keys, find properties that match the sequence. The result is a list of property
        because there could be more than one path that matches the key sequence.
        """
        if len(payload_keys) == 0:
            raise ValueError('no payload keys to find for properties')

        candidate_properties = body.candidate_properties()
        while len(payload_keys) > 0:
            matched_properties = []
            for candidate_property in candidate_properties:
                if not isinstance(candidate_property, Variable):
                    continue
                if candidate_property.match_key(payload_keys[0]):
                    matched_properties.append(candidate_property)

            if len(matched_properties) == 0:  # No match is found
                return []

            # match is found
            payload_keys.pop(0)
            if len(payload_keys) == 0:
                return [p for p in matched_properties if isinstance(p, Variable)]

            # Prepare for the next round of key matching
            candidate_properties = []
            for matched_property in matched_properties:
                candidate_properties.extend(self._sub_properties_for_property(self.types, matched_property))

        return []  # should not reach here, but for safe

    def _sub_properties_for_keys(self, body: Body, payload_keys: List[str]) -> List[Variable]:
        """
        For the given payload_keys, find properties that match the sequence, then return their sub-properties.
        """
        if len(payload_keys) == 0:
            return body.candidate_properties()
        else:
            matched_properties = self._properties_for_keys(body, payload_keys)
            sub_properties = []
            for matched_property in matched_properties:
                sub_properties.extend(self._sub_properties_for_property(self.types, matched_property))
            return sub_properties

    def _build_common_params(self) -> Dict[str, List[str]]:
        type_definition = self.types[TypeName('CommonQueryParameters', '_spec_utils')]
        if not isinstance(type_definition, Interface):
            _logger.warning(f'CommonQueryParameters type kind [{type_definition.kind}] unprocessable')
            return {}
        common_parameters = {}
        for prop in type_definition.properties:
            query_parameter = Variable.from_dict(prop)
            common_parameters[query_parameter.name] = self._filter_for_param_values(
                query_parameter.candidate_values(self.types)
            )
        return common_parameters

    @staticmethod
    def _sub_properties_for_property(types: Types, matched_property: Variable):
        try:
            if isinstance(matched_property.value, ArrayOf):
                # penetrate array
                return matched_property.value.get_member().candidate_properties(types)
            else:
                return matched_property.candidate_properties(types)
        except KeyError as e:
            _logger.debug(f'error in _sub_properties_for_property: {e}')
            return []

    @staticmethod
    def _can_match(ts, ps):
        """
        Test whether the input path (ts) can match the candidate path (ps).
        The rule is basically a placeholder can match any string other than
        the ones leading with underscore.
        """
        for t, p in zip(ts, ps):
            if t != p:
                if t.startswith('_'):
                    return False
                if not (p.startswith('{') and p.endswith('}')):
                    return False
        return True

    @staticmethod
    def _filter_for_param_values(candidate_values) -> List[str]:
        final_values = []
        for value in candidate_values:
            if value is True:
                final_values.append('true')
            elif value is False:
                final_values.append('false')
            elif value is None:
                final_values.append('null')
            elif isinstance(value, numbers.Number):
                final_values.append(str(value))
            elif isinstance(value, str):
                final_values.append(value)
            else:
                _logger.debug(f'non-usable query parameter value: {value!r}')
        return final_values

    @staticmethod
    def _request_has_common_query_params(request: Request) -> bool:
        attached_behaviours = request.get_attached_behaviours()
        return attached_behaviours is not None and 'CommonQueryParameters' in attached_behaviours

    @staticmethod
    def _get_query_parameter(param_name: str, request: Request) -> Union[Variable, None]:
        query: Dict = request.get_query()
        if query is None:
            return None
        d = query.get(param_name, None)
        if d is None:
            return None
        return Variable.from_dict(d)



import functools
import itertools
import logging
import os
import re
from typing import Iterable
from unittest.mock import MagicMock

from configobj import ConfigObj

from peek.ast import BinOpNode, DictNode, NameNode, Node, SymbolNode, TextNode
from peek.config import config_location
from peek.parser import PeekParser
from peek.visitors import FormattingVisitor, Ref
from peek.vm import _BIN_OP_FUNCS, PeekVM

_CONST_SIMPLE_PATTERN = re.compile(r'^(export )?const (?P<name>\w+)[^=]* = (?P<rest>{[^;]*);?')
_CONST_COMPLEX_PATTERN = re.compile(
    r'^(export )?const (?P<name>\w+)[^=]* = ' r'\(specService: SpecDefinitionsService\) => (?P<rest>{[^;]*);?'
)
_CONST_MULTI_PATTERN = re.compile(r'^(export )?const {(?P<names>[^}]*)} = (?P<value>.+);')
_RECORD_MEMBER_PATTERN = re.compile(r'(?P<assign>^\w+\.\w+ = {[^;]*);?')
_FUNCTION_PATTERN = re.compile(r'function \((?P<args>[^)]*)\) { *;?]')
_RETURN_PATTERN = re.compile(r'return (?P<value>.*)')

_logger = logging.getLogger(__name__)


# TODO: ingest.ts is parsed but processors seem to be ignored. Also don't know how it is referenced.


def build_js_specs(kibana_dir, use_cache_file):
    # Cache is not for efficiency, but rather because the spec building from TypeScript
    # files is hacky and likely to go wrong with new Kibana releases. Cache it so at least
    # we can have some usable version till a fix is ready for new releases.
    cached_extended_specs_file = os.path.expanduser(config_location()) + 'extended_specs.es'
    source = None
    if use_cache_file and os.path.exists(cached_extended_specs_file):
        _logger.info(f'Found cached extended specs file: {cached_extended_specs_file!r}')
        with open(cached_extended_specs_file) as ins:
            source = ins.read()
    spec_parser = JsSpecParser(kibana_dir, source=source)
    nodes = spec_parser.parse()
    spec_evaluator = JsSpecEvaluator()
    specs = spec_evaluator.visit(nodes)
    if use_cache_file and source is None:
        spec_parser.save(cached_extended_specs_file)
    _logger.info('Complete building extended specs')
    return specs


class JsSpecParser:
    def __init__(self, kibana_dir, source=None):
        self.kibana_dir = kibana_dir
        self.source = source
        self.nodes = []

    def parse(self):
        if self.source is None:
            self.source = self._extract_all()
        parser = PeekParser()
        self.nodes = parser.parse(self.source, log_level='WARNING')
        return self.nodes

    def save(self, output_file):
        content = []
        for node in self.nodes:
            content.append(FormattingVisitor(pretty=True).visit(node))
        with open(output_file, 'w') as outs:
            outs.write('\n'.join(content))

    def _extract_all(self):
        spec_file_contents = self.load_ts_specs(self.kibana_dir)
        sources = []
        # TODO: this file is imported by oss/query/dsl, we manually prioritize it but automation would be good
        dependency_file = '/oss/query/templates'
        if dependency_file in spec_file_contents:
            sources.extend(self._extract_from_one_file(dependency_file, spec_file_contents.pop(dependency_file)))
        for file_name, file_content in spec_file_contents.items():
            sources.extend(self._extract_from_one_file(file_name, file_content))
        text = '\n'.join(sources)
        return text.replace("rules['*']", "rules.'*'")

    def _extract_from_one_file(self, file_name, file_content):
        sources = [f'// {file_name}']
        state = ''
        for line in file_content.splitlines():
            stripped = line.strip()
            if '...[' in stripped:
                stripped = stripped.replace('...[', '_.splat % [')
            elif '...' in stripped:
                stripped = stripped.replace('...', '"...": @')
            if state == 'comments':
                if stripped.startswith('*/'):
                    state = ''
                else:
                    continue
            elif state == 'import':
                if stripped.endswith(';'):
                    state = ''
                continue
            elif stripped.startswith('/*'):
                if not stripped.endswith('*/'):
                    state = 'comments'
                continue
            elif stripped == '' or stripped.startswith('// '):
                continue
            elif stripped.startswith('import '):
                if stripped.endswith(';'):
                    continue
                elif stripped.endswith('{'):
                    assert state == '', f'import within other state: {state!r}'
                    state = 'import'
                    continue
                else:
                    raise ValueError(f'Unrecognised import variant: {stripped!r}')
            elif self._try_speical_0001(sources, stripped):
                continue
            elif self._try_special_0002(sources, stripped):
                continue
            elif self._try_const_simple_pattern(sources, stripped):
                continue
            elif self._try_const_complex_pattern(sources, stripped):
                continue
            elif self._try_const_multi_pattern(sources, stripped):
                continue
            elif self._try_record_member_assignment(sources, stripped):
                continue
            elif self._try_spec_service(sources, stripped):
                continue
            elif self._try_function(sources, stripped):
                continue
            elif self._try_return(sources, stripped):
                continue
            elif stripped.startswith('export '):
                continue
            elif stripped == 'gap_policy,':
                sources.append('"gap_policy": gap_policy,')
            elif stripped.endswith(';'):
                sources.append(stripped[:-1])
            else:
                sources.append(stripped)
        return sources

    def _try_const_simple_pattern(self, sources, stripped):
        m = _CONST_SIMPLE_PATTERN.match(stripped)
        if m is not None:
            sources.append(f'let {m["name"]} = {m["rest"]}')
            return True
        else:
            return False

    def _try_const_complex_pattern(self, sources, stripped):
        m = _CONST_COMPLEX_PATTERN.match(stripped)
        if m is not None:
            sources.append(f'for {m["name"]} in [0] {m["rest"]}')
            return True
        else:
            return False

    def _try_const_multi_pattern(self, sources, stripped):
        m = _CONST_MULTI_PATTERN.match(stripped)
        if m is not None:
            for name in m['names'].split(','):
                sources.append(f'let {name.strip()} = {m["value"]}')
            return True
        else:
            return False

    def _try_record_member_assignment(self, sources, stripped):
        m = _RECORD_MEMBER_PATTERN.match(stripped)
        if m is not None:
            sources.append(f'let {m["assign"]}')
            return True
        else:
            return False

    def _try_spec_service(self, sources, stripped):
        if stripped.startswith('specService.'):
            sources.append(self._trim_semicolon(stripped.replace('specService.', 'specService ', 1)))
            return True
        else:
            return False

    def _try_function(self, sources, stripped):
        if stripped == 'function (s) {':
            sources.append('{')
            return True
        elif 'function (s) {' in stripped:
            sources.append(stripped.replace('function (s) {', '{'))
            return True
        else:
            return False

    def _try_return(self, sources, stripped):
        m = _RETURN_PATTERN.match(stripped)
        if m is not None:
            value = m["value"]
            if value.endswith(';'):
                value = value[:-1]
            sources.append(f'"return": """{value}"""')
            return True
        else:
            return False

    def _try_speical_0001(self, sources, stripped):
        if stripped == ''''*': { terms, histogram, date_histogram },''':
            sources.append('"*": { "terms": terms, "histogram": histogram, "date_histogram": date_histogram },')
            return True
        else:
            return False

    def _try_special_0002(self, sources, stripped):
        # TODO: filters.m is not handled
        if stripped == '''filters.m = filters.missing = {''':
            sources.append('let filters.missing = {')
            return True
        else:
            return False

    def load_ts_specs(self, kibana_dir):
        oss_path = os.path.join(kibana_dir, 'src', 'plugins', 'console', 'server', 'lib', 'spec_definitions', 'js')
        xpack_path = os.path.join(
            kibana_dir, 'x-pack', 'plugins', 'console_extensions', 'server', 'lib', 'spec_definitions', 'js'
        )

        spec_file_contents = self._load_ts_specs(oss_path)
        spec_file_contents.update(self._load_ts_specs(xpack_path))
        return spec_file_contents

    def _load_ts_specs(self, base_dir):
        prefix = 'x-pack' if 'x-pack' in base_dir else 'oss'
        spec_file_contents = {}
        for root, dirs, files in os.walk(base_dir):
            for f in files:
                if not f.endswith('.ts'):
                    continue
                if f in ('shared.ts', 'index.ts'):  # TODO: skip
                    continue
                with open(os.path.join(root, f)) as ins:
                    file_key = os.path.join('/', prefix, root[len(base_dir) + 1 :], f[:-3])
                    spec_file_contents[f'{file_key}'] = ins.read()

        return spec_file_contents

    def _trim_semicolon(self, line):
        return line[:-1] if line.endswith(';') else line


mock_app = MagicMock()
ConfigObj(
    {
        'load_extension': False,
    }
)
mock_app.display = MagicMock()
mock_app.parser = PeekParser()


class JsSpecEvaluator(PeekVM):
    def __init__(self):
        def flexible_dot(left_operand, right_operand):
            if isinstance(left_operand, list) and right_operand == 'sort':
                return lambda app: _sort(left_operand)
            elif isinstance(left_operand, list) and right_operand == 'flatMap':
                return lambda app, func: functools.partial(_flat_map, left_operand)(app, func)
            else:
                return _BIN_OP_FUNCS['.'](left_operand, right_operand)

        def flexible_mod(left_operand, right_operand):
            if left_operand == self.builtins['_']['splat']:
                return left_operand(right_operand)
            else:
                return _BIN_OP_FUNCS['%'](left_operand, right_operand)

        bin_op_funcs = dict(_BIN_OP_FUNCS)
        bin_op_funcs.update(
            {
                '.': flexible_dot,
                '%': flexible_mod,
            }
        )

        super().__init__(mock_app, bin_op_funcs=bin_op_funcs)
        mock_app.vm = self
        self.builtins = {
            '_': {
                'defaults': _defaults,
                'flatten': _flatten,
                'map': _map,
                'splat': self._splat,
            },
            'return': lambda _, v: v,
            'specService': lambda _, v: v,
            'addGlobalAutocompleteRules': add_global_autocomplete_rules,
            'addEndpointDescription': add_endpoint_description,
        }
        self.context = {
            'BOOLEAN': {"__one_of": [True, False]},
            'GLOBAL': {},
        }

    def visit(self, nodes):
        for node in nodes:
            self.execute_node(node)
        # Remove all loop iterator variables
        self.context = {k: v for k, v in self.context.items() if v != 0 and k != 's'}
        return self.context

    def visit_bin_op_node(self, node: BinOpNode):
        if node.op_token.value == '.' and isinstance(node.right_node, NameNode):
            node.right_node = TextNode(node.right_node.token)
        super(JsSpecEvaluator, self).visit_bin_op_node(node)

    def visit_dict_node(self, node: DictNode):
        d_ref = Ref()
        with self.consumer(lambda v: d_ref.set(v)):
            self._do_visit_dict_node(node, resolve_key_name=False)
        d = d_ref.get()
        if d.get('...', None) is not None:
            splat = d.pop('...')
            assert isinstance(splat, dict), f'Cannot splat type other than dict, got: {splat!r}'
            d.update(splat)
        self.consume(d)

    def visit_symbol_node(self, node: SymbolNode):
        self.consume(self.get_value(node.token.value))

    def visit_es_api_call_node(self, node):
        raise ValueError(f'SpecEvaluator cannot take ES API call node: {node!r}')

    def visit_shell_out_node(self, node):
        raise ValueError(f'SpecEvaluator cannot take shell out node: {node!r}')

    def _unwind_lhs(self, node: Node):
        if isinstance(node, BinOpNode):
            if node.op_token.value == '.' and isinstance(node.right_node, NameNode):
                node.right_node = TextNode(node.right_node.token)
        super(JsSpecEvaluator, self)._unwind_lhs(node)

    def _splat(self, values: Iterable):
        if not isinstance(values, Iterable):
            raise ValueError(f'Expect Iterable, got {values!r}')
        for v in values[:-1]:
            self.consume(v)
        return values[-1]


def add_global_autocomplete_rules(app, name, rule):
    app.vm.context['GLOBAL'][name] = rule


def add_endpoint_description(app, name, rule):
    app.vm.context[name] = rule


def _map(app, values, ret):
    ret_nodes = app.parser.parse('return ' + ret['return'], log_level='WARNING')
    results = []
    for value in values:
        app.vm.context['s'] = value
        v_ref = Ref()
        with app.vm.consumer(lambda v: v_ref.set(v)):
            ret_nodes[0].args_node.value_nodes[0].accept(app.vm)
        results.append(v_ref.get())
    return results


def _defaults(app, *args):
    d = {}
    for arg in args:
        d.update(arg)
    return d


def _flatten(app, *args):
    """
    This is specific to mappings.properties.format. NOT a generic flatten function
    """
    return list(itertools.chain(*args[0][0]))


def _flat_map(values, app, func):
    results = _map(app, values, func)
    return list(itertools.chain(*results))


def _sort(values):
    return sorted(values)



import json
import logging
import os
import random

from configobj import ConfigObj

from peek import __version__
from peek.common import DEFAULT_SAVE_NAME
from peek.config import config_location, get_global_config
from peek.connection import ConnectFunc, EsClientManager
from peek.display import PeekEncoder
from peek.errors import PeekError
from peek.krb import KrbAuthenticateFunc
from peek.oidc import OidcAuthenticateFunc
from peek.saml import SamlAuthenticateFunc

_logger = logging.getLogger(__name__)


class ConfigFunc:
    def __call__(self, app, **options):
        if not options:
            return {'location': config_location(), 'config': app.config}

        extra_config = {}
        for key, value in options.items():
            parent = extra_config
            key_components = key.split('.')
            for key_component in key_components[:-1]:
                child = parent.get(key_component)
                if child is None:
                    parent[key_component] = {}
                elif not isinstance(child, dict):
                    _logger.warning(
                        f'Config key {key!r} conflicts. '
                        f'Value of {key_component!r} is not a dict, '
                        f'but {type(child)!r}'
                    )
                    parent = None
                    break
                parent = parent[key_component]

            if isinstance(parent, dict):
                parent[key_components[-1]] = value

        # TODO: saner merge that does not change data type, e.g. from dict to primitive and vice versa
        app.config.merge(ConfigObj(extra_config))

    @property
    def description(self):
        return 'View and set config options'


class ConnectionFunc:
    def __call__(self, app, current=None, **options):
        options = consolidate_options(
            options,
            {
                'info': app.es_client_manager.index_current,
                'remove': app.es_client_manager.index_current,
                'keep': app.es_client_manager.index_current,
                'wipe': True,
            },
        )

        current = current if current is not None else options.get('current', None)
        if current is not None:
            app.es_client_manager.set_current(current)

        wipe = options.get('wipe', None)
        if wipe:
            for _ in range(len(app.es_client_manager.clients())):
                app.es_client_manager.remove_client()

        remove = options.get('remove', None)
        if remove is not None:
            app.es_client_manager.remove_client(remove)

        move = options.get('move', None)
        if move is not None:
            app.es_client_manager.move_current_to(move)

        keep = options.get('keep', None)
        if keep is not None:
            app.es_client_manager.keep_client(keep)

        rename = options.get('rename', None)
        if rename:
            app.es_client_manager.current.name = str(rename)

        info = options.get('info', None)
        if info is not None:
            return app.es_client_manager.get_client(info).info()

        return str(app.es_client_manager)

    @property
    def options(self):
        return {
            'current': None,
            'remove': None,
            'move': None,
            'rename': None,
            'info': None,
            'keep': None,
            '@info': None,
            '@remove': None,
            '@keep': None,
            '@wipe': None,
        }

    @property
    def description(self):
        return 'List connections and set current connection'


class SessionFunc:
    def __call__(self, app, **options):
        options = consolidate_options(
            options,
            {
                'load': DEFAULT_SAVE_NAME,
                'save': DEFAULT_SAVE_NAME,
                'clear': None,
            },
        )

        if not options:
            return app.history.list_sessions()

        if 'load' in options:
            load = options.get('load')
            data = app.history.load_session(load)
            if data is None:
                raise PeekError(f'Session not found: {load!r}')
            else:
                app.es_client_manager = EsClientManager.from_dict(app, json.loads(data))
            return str(app.es_client_manager)

        elif 'save' in options:
            save = options.get('save')
            app.history.save_session(save, json.dumps(app.es_client_manager.to_dict()))
            return f'Session save as: {save!r}'

        elif 'remove' in options:
            remove = options.get('remove')
            if app.history.delete_session(remove):
                return f'Session removed: {remove!r}'
            else:
                raise PeekError(f'Session not found: {remove!r}')

        elif 'clear' in options:
            app.history.clear_sessions()
            return 'All sessions cleared'
        else:
            raise PeekError(f'Unknown options: {options}')

    @property
    def options(self):
        return {
            'save': None,
            'load': None,
            'remove': None,
            '@save': DEFAULT_SAVE_NAME,
            '@load': DEFAULT_SAVE_NAME,
            '@clear': None,
        }

    @property
    def description(self):
        return 'Session management - save and load sessions to quickly get back to an known state'


class RunFunc:
    def __call__(self, app, file, **options):
        should_echo = options.get('echo', True)
        is_capture = options.get('is_capture')
        if is_capture:
            # preprocess to remove non-executable sections of captured output
            with open(file) as ins:
                filtered = self._filter_captured_output(ins.readlines())
            app.process_input('\n'.join(filtered), echo=should_echo)
        else:
            with open(file) as ins:
                app.process_input(ins.read(), echo=should_echo)

    @property
    def options(self):
        return {'echo': True, 'is_capture': False}

    @property
    def description(self):
        return 'Load and execute external script'

    def _filter_captured_output(self, lines):
        filtered = []
        state = 0
        while len(lines) > 0:
            line = lines.pop(0).strip()
            if state == 0:
                if line == '===' or line.startswith('==='):
                    state = 1
                elif line == 'capture' or line.startswith('capture '):
                    continue
                elif line == '>>>' or line.startswith('>>> '):
                    continue
                else:
                    filtered.append(line)
            elif state == 1:
                if line == '>>>' or line.startswith('>>> '):
                    state = 0
                else:
                    continue

        return filtered


class HistoryFunc:
    def __call__(self, app, index=None, **options):
        if index is None:
            history = []
            for entry in app.history.load_recent(size=options.get('size', 100)):
                history.append(f'{entry[0]:>6} {entry[1]!r}')
            return '\n'.join(history)
        else:
            entry = app.history.get_entry(index)
            if entry is None:
                raise PeekError(f'History not found for index: {index}')
            app.process_input(entry[1])

    @property
    def options(self):
        return {'size': 100}

    @property
    def description(self):
        return 'View history and execute by history index'


class RangeFunc:
    def __call__(self, app, start, stop, step=1):
        return list(range(start, stop, step))

    @property
    def description(self):
        return 'Range over given start and stop (exclusive) with an optional step'


class RandIntFunc:
    def __call__(self, app, min_value=None, max_value=None):
        if min_value is None and max_value is None:
            min_value = 0
            max_value = 100
        elif max_value is None:
            max_value = min_value
            min_value = 0
        return random.randrange(min_value, max_value)

    @property
    def description(self):
        return 'Return a random integer between the given [min, max) values, default to [0, 100)'


class EchoFunc:
    def __init__(self):
        self.builtin_lookup = {
            True: 'true',
            False: 'false',
            None: 'null',
        }

    def __call__(self, app, *args, **options):
        self.function_lookup = {v: k for k, v in app.vm.functions.items()}
        content = []
        for arg in args:
            content.append(json.dumps(arg, cls=PeekEncoder, app=app, separators=(',', ':')))

        content = ' '.join(content)
        if 'file' in options:
            end = '\n'  # this does not make sense in interactive mode, so hardcode it for now
            fname = options.get('file')
            if not isinstance(fname, str):
                raise PeekError(f'file name must be string, got {fname!r}')
            with open(fname, 'a') as outs:
                outs.write(f'{content}{end}')
        else:
            return content

    @property
    def options(self):
        return {'file': None}

    @property
    def description(self):
        return 'Print given items in JSON format, optionally appending to a file'


class CaptureFunc:
    def __call__(self, app, f=None, **options):
        directives = options.get('@')
        if not directives:
            return app.capture.status()

        # Only honor first directive
        directive = directives[0]
        if directive == 'start':
            return app.start_capture(f)

        elif directive == 'stop':
            return app.stop_capture()
        else:
            raise PeekError(f'Unknown capture directive: {directive}')

    @property
    def options(self):
        return {'@start': None, '@stop': None}

    @property
    def description(self):
        return 'Capture session IO into a file'


class GetEnvFunc:
    def __call__(self, app, name):
        return os.getenv(name, '')

    @property
    def description(self):
        return (
            'Get value of the environmental variable of the given name. '
            'Returns an empty string if name does not exist'
        )


class ResetFunc:
    def __call__(self, app, **options):
        app.reset()

    @property
    def description(self):
        return 'Reset current session back to its initial start state'


class ExitFunc:
    def __call__(self, app):
        if not app.batch_mode:
            app.signal_exit()

    @property
    def description(self):
        return 'Exit the current interactive session'


class HelpFunc:
    def __call__(self, app, func=None, **options):
        if func is None:
            return '\n'.join(f'{k} - {getattr(v, "description", "")}' for k, v in app.vm.functions.items())

        for k, v in app.vm.functions.items():
            if v == func:
                description = getattr(v, "description", None)
                header = k + (f' - {description}' if description else '')
                return f'{header}\n{getattr(func, "options", {})}'
        else:
            raise PeekError(f'No such function: {func}')

    @property
    def description(self):
        return 'List available functions and show help message of a function'


class VersionFunc:
    def __call__(self, app):
        import elastic_transport

        return f'"Peek (v{__version__})"\n"elastic_transport (v{elastic_transport.__version__})"'

    @property
    def description(self):
        return 'Show version info of the installation'


class DownloadApiSpecsFunc:
    def __call__(self, app, **options):
        config_dir = config_location()
        if not os.path.exists(config_dir):
            raise FileNotFoundError(
                f'Config directory does not exist: {config_dir!r}. '
                f'Please create it before downloading API spec files'
            )

        if app.config.as_bool('prefer_elasticsearch_specification'):
            schema_filepath = os.path.join(config_dir, 'schema.json')
            if os.path.exists(schema_filepath):
                raise RuntimeError(f'schema file already exists [{schema_filepath}]. Please remove it before download.')
            git_branch = options.get('version', '8.12')
            import urllib.request

            url = (
                f'https://raw.githubusercontent.com/elastic/elasticsearch-specification/'
                f'{git_branch}/output/schema/schema.json'
            )
            data = urllib.request.urlopen(url).read()
            with open(schema_filepath, 'wb') as outs:
                outs.write(data)
            app.completer.init_api_completer()
            return f'Elasticsearch specification [{git_branch}] downloaded and initialized'
        else:
            existing_kibana_dirs = [
                os.path.join(config_dir, d) for d in os.listdir(config_dir) if d.startswith('kibana-')
            ]
            if existing_kibana_dirs:
                raise RuntimeError(
                    f'Existing {"directory" if len(existing_kibana_dirs) == 1 else "directories"} '
                    f'found for API specs: {existing_kibana_dirs}. '
                    f'Please remove {"it" if len(existing_kibana_dirs) == 1 else "them"} '
                    f'before download new spec files.'
                )

            import io
            import urllib.request
            import zipfile

            kibana_version = options.get('version', '7.9.1')
            kibana_release_url = f'https://github.com/elastic/kibana/archive/v{kibana_version}.zip'
            app.display.info(f'Downloading from {kibana_release_url} ... This may take a few minutes ...')
            data = urllib.request.urlopen(kibana_release_url).read()
            zf = zipfile.ZipFile(io.BytesIO(data))
            for info in zf.infolist():
                if 'spec_definitions' in info.filename:
                    zf.extract(info, path=config_dir)
            app.completer.init_api_completer()
            return f'Version {kibana_version} API spec files are ready'

    @property
    def options(self):
        if get_global_config().as_bool('prefer_elasticsearch_specification'):
            return {'version': '8.12'}
        else:
            return {'version': '7.9.1'}

    @property
    def description(self):
        if get_global_config().as_bool('prefer_elasticsearch_specification'):
            return 'Download and reload API spec file from the Elasticsearch Specification project.'
        else:
            return (
                'Download and reload Elasticsearch API spec files from the Kibana project. '
                'It may takes a few minutes depending on the network speed.'
            )


class DownloadExtensionFileFunc:

    def __call__(self, app, url=None, **options):
        if not url:
            raise PeekError('URL of the extension file must be provided')

        default_extension_path = os.path.join(config_location(), 'extensions')
        if not os.path.exists(default_extension_path):
            os.makedirs(default_extension_path, exist_ok=True)

        filename = options.get('filename', None)
        if filename is None:
            from urllib3.util import parse_url
            parsed = parse_url(url)
            filename = os.path.basename(parsed.path)

        if not filename.endswith('.py'):
            filename += '.py'
        filename = os.path.join(default_extension_path, filename)

        import urllib.request
        data = urllib.request.urlopen(url).read()
        with open(filename, 'wb') as outs:
            outs.write(data)

        app.vm._load_one_extension_path(default_extension_path)
        return f'Extension file [{filename}] downloaded and initialized'

    @property
    def description(self):
        return 'Download an extension file from the specified URL and save it to the default extension directory'


class ContextFunc:

    def __call__(self, app, **options):
        keys = sorted([k for k, v in app.vm.context.items() if not callable(v)])
        lines = []
        for key in keys:
            lines.append(f'{key} = {app.vm.context[key]!r}')
        return '\n'.join(lines)


def consolidate_options(options, defaults):
    """
    Merge shorthanded @symbol into normal options kv pair with provided defaults
    """
    final_options = {k: v for k, v in options.items() if k != '@'}
    for symbol in options.get('@', []):
        if symbol not in defaults:
            raise PeekError(f'Unknown shorthanded flag: {symbol}')
        final_options[symbol] = defaults[symbol]
    return final_options


EXPORTS = {
    'connect': ConnectFunc(),
    'config': ConfigFunc(),
    'connection': ConnectionFunc(),
    'session': SessionFunc(),
    'run': RunFunc(),
    'history': HistoryFunc(),
    'echo': EchoFunc(),
    'range': RangeFunc(),
    'randint': RandIntFunc(),
    'capture': CaptureFunc(),
    'getenv': GetEnvFunc(),
    'reset': ResetFunc(),
    'exit': ExitFunc(),
    'help': HelpFunc(),
    'version': VersionFunc(),
    '_download_api_specs': DownloadApiSpecsFunc(),
    'saml_authenticate': SamlAuthenticateFunc(),
    'oidc_authenticate': OidcAuthenticateFunc(),
    'krb_authenticate': KrbAuthenticateFunc(),
    '_download_extension_file': DownloadExtensionFileFunc(),
    'context': ContextFunc(),
}



import logging
import os
import platform
from os.path import dirname, expanduser
from typing import Iterable

from configobj import ConfigObj

_logger = logging.getLogger(__name__)


def config_location():
    from peek import __package__ as package_name

    if 'XDG_CONFIG_HOME' in os.environ:
        return f'%s/{package_name}/' % expanduser(os.environ['XDG_CONFIG_HOME'])
    elif platform.system() == 'Windows':
        return os.getenv('USERPROFILE') + f'\\AppData\\Local\\{package_name}\\'
    else:
        return expanduser(f'~/.config/{package_name}/')


def ensure_dir_exists(path):
    parent_dir = expanduser(dirname(path))
    os.makedirs(parent_dir, exist_ok=True)


def load_config(package_config_file: str, config_file: str = None, extra_config_options: Iterable[str] = None):
    config = ConfigObj(package_config_file)
    if config_file is not None:
        config.merge(ConfigObj(config_file))

    if extra_config_options:
        extra_config = {}
        for config_option in extra_config_options:
            parent = extra_config
            key, value = config_option.split('=', 1)
            key_components = key.split('.')
            for key_component in key_components[:-1]:
                child = parent.get(key_component)
                if child is None:
                    parent[key_component] = {}
                elif not isinstance(child, dict):
                    _logger.warning(
                        f'Config key [{key}] conflicts. '
                        f'Value of [{key_component}] is not a [dict], '
                        f'but [{type(child)}]'
                    )
                    parent = None
                    break
                parent = parent[key_component]

            if isinstance(parent, dict):
                parent[key_components[-1]] = value

        config.merge(ConfigObj(extra_config))

    return config


_CONFIG = None


def get_config(config_file: str = None, extra_config_options: Iterable[str] = None):
    from peek import __file__ as package_root

    package_root = os.path.dirname(package_root)
    package_config_file = os.path.join(package_root, 'peekrc')
    default_config_file = expanduser(config_location() + 'peekrc')
    ensure_dir_exists(default_config_file)
    config_file = config_file or default_config_file
    global _CONFIG
    _CONFIG = load_config(package_config_file, config_file, extra_config_options)
    return _CONFIG


def get_global_config():
    if _CONFIG is None:
        raise ValueError("global config not initialized")
    return _CONFIG



import ast
import json
import logging
from typing import Dict, List

from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.document import Document
from pygments.token import Name, String

from peek.common import PeekToken
from peek.es_api_spec.api_completer import ESApiCompleter
from peek.es_api_spec.kspec_js import build_js_specs
from peek.es_api_spec.kspec_json import load_json_specs
from peek.lexers import EOF, Assign, BracketLeft, Colon, Comma, CurlyLeft, CurlyRight, DictKey, PathPart, Slash
from peek.parser import ParserEvent, ParserEventType

_logger = logging.getLogger(__name__)


class KibanaSpecESApiCompleter(ESApiCompleter):
    def __init__(self, app, kibana_dir):
        self.app = app
        self.kibana_dir = kibana_dir
        self.specs = self._build_specs()

    def complete_url_path(self, document: Document, complete_event: CompleteEvent, method, path_tokens):
        cursor_token = path_tokens[-1]
        _logger.debug(f'Completing URL path: {cursor_token}')
        token_stream = [t.value for t in path_tokens if t.ttype is not Slash]
        if cursor_token.ttype is PathPart:
            token_stream.pop()
        candidates = []
        for api_name, api_spec in self.specs.items():
            if 'methods' not in api_spec:
                continue
            if method not in api_spec['methods']:
                continue
            for api_path in api_spec['patterns']:
                ps = [p for p in api_path.split('/') if p]
                # Nothing to complete if the candidate is shorter than current input
                if len(token_stream) >= len(ps):
                    continue
                if not can_match(token_stream, ps):
                    continue
                candidate = '/'.join(ps[len(token_stream) :])
                candidates.append(Completion(candidate))
        return candidates

    def complete_query_param_name(self, document: Document, complete_event: CompleteEvent, method, path_tokens):
        _logger.debug(f'Completing URL query param name: {path_tokens[-1]}')
        token_stream = [t.value for t in path_tokens if t.ttype is PathPart]
        candidates = set()
        for api_spec in matchable_specs(method, token_stream, self.specs):
            candidates.update(api_spec['url_params'].keys())
        return [Completion(c) for c in candidates]

    def complete_query_param_value(self, document: Document, complete_event: CompleteEvent, method, path_tokens):
        _logger.debug(f'Completing URL query param value: {path_tokens[-1]}')
        param_name_token = path_tokens[-2] if path_tokens[-1].ttype is Assign else path_tokens[-3]
        token_stream = [t.value for t in path_tokens if t.ttype is PathPart]
        candidates = set()
        for api_spec in matchable_specs(method, token_stream, self.specs):
            v = api_spec['url_params'].get(param_name_token.value, None)
            if v is None:
                continue
            if v == '__flag__':
                candidates.update(('true', 'false'))
            elif isinstance(v, list) and len(v) > 0:
                candidates.update(v)
        return [Completion(c) for c in candidates]

    # TODO: refactor with parser state tracker
    def complete_payload(self, document, complete_event, method, path_tokens, payload_tokens, payload_events):
        _logger.debug(f'Completing for API payload: {method!r} {path_tokens!r} {payload_tokens!r}')
        rules = self._find_rules_for_method_and_url_path(method, path_tokens)
        if rules is None:
            return [], {}

        payload_keys = []
        curly_level = 0
        for t in payload_tokens[:-1]:
            if t.ttype is CurlyLeft:
                curly_level += 1
            elif t.ttype is CurlyRight:
                curly_level -= 1
                payload_keys.pop()
            elif t.ttype is DictKey:
                if len(payload_keys) == curly_level - 1:
                    payload_keys.append(ast.literal_eval(t.value))
                elif len(payload_keys) == curly_level:
                    payload_keys[-1] = ast.literal_eval(t.value)
                else:
                    raise ValueError(f'Error when counting curly level {curly_level} and keys {payload_keys}')

        _logger.debug(f'Payload status: level: {curly_level}, keys: {payload_keys}')
        if curly_level == 0:  # not even in the first curly bracket, no completion
            return [], {}

        # Remove the payload key that is at the same level
        if curly_level == len(payload_keys):
            payload_keys.pop()

        rules = self._resolve_rules_for_keys(rules, payload_keys)
        if rules is None:
            _logger.debug(f'Rules not available for key: {payload_keys!r}')
            return [], {}
        _logger.debug(f'Found rules for payload keys: {rules!r}')
        rules = dict(rules)  # avoid mutating original rules

        candidates = []
        for rule_key in list(rules.keys()):
            if rule_key == '__scope_link':
                _logger.debug(f'__scope_link not processed for rules: {rules}')
            elif rule_key == '__one_of':
                _logger.debug(f'__one_of not processed for rules: {rules}')
            elif rule_key == '*':
                _logger.debug(f'Wildcard * not processed for rules: {rules}')
            elif rule_key == '__template':
                __template = rules.get('__template', None)
                rules.pop('__template')
                if isinstance(__template, dict):
                    for k, v in __template.items():
                        if k not in rules:
                            rules[k] = v
                            candidates.append(Completion(k))
            else:
                candidates.append(Completion(rule_key))

        return candidates, rules

    def complete_payload_value(
        self,
        document: Document,
        complete_event: CompleteEvent,
        method: str,
        path_tokens: List[PeekToken],
        payload_tokens: List[PeekToken],
        payload_events: List[ParserEvent],
    ):
        _logger.debug(f'Completing for API payload value: {method!r} {path_tokens!r} {payload_tokens!r}')
        rules = self._find_rules_for_method_and_url_path(method, path_tokens)
        if rules is None:
            return [], {}

        unpaired_dict_key_tokens = []
        for payload_event in payload_events:
            if payload_event.type is ParserEventType.BEFORE_DICT_KEY_EXPR:
                _logger.debug(f'No completion is possible for payload with dict key expr: {payload_event.token}')
                return [], {}
            elif payload_event.type is ParserEventType.DICT_KEY:
                unpaired_dict_key_tokens.append(payload_event.token)
            elif payload_event.type is ParserEventType.AFTER_DICT_VALUE and payload_event.token.ttype is not EOF:
                unpaired_dict_key_tokens.pop()

        if not unpaired_dict_key_tokens:  # should not happen
            _logger.warning('No unpaired dict key tokens are found')
            return [], {}

        payload_keys = [ast.literal_eval(t.value) for t in unpaired_dict_key_tokens]
        _logger.debug(f'Payload keys are: {payload_keys}')

        rules = self._resolve_rules_for_keys(rules, payload_keys, unwrap_value_for_last_key=False)
        if rules is None:
            _logger.debug(f'Rules not available for keys: {payload_keys!r}')
            return [], {}
        _logger.debug(f'Found rules for payload keys: {rules!r}')

        # Find last Colon position
        for index_colon in range(len(payload_tokens) - 1, -1, -1):
            if payload_tokens[index_colon].ttype is Colon:
                break
        else:
            _logger.warning(f'Should not happen - Colon not found in payload: {payload_tokens}')
            return [], {}

        if index_colon == len(payload_tokens) - 1:  # Colon is the last token
            _logger.debug('Colon is the last token')
            # The simpler case when value position has nothing yet
            if isinstance(rules, dict):
                if '__one_of' in rules:
                    return [Completion('""' if isinstance(c, str) else json.dumps(c)) for c in rules['__one_of']], rules
                else:
                    return [Completion('{}')], rules
            elif isinstance(rules, list):
                if rules and isinstance(rules[0], dict):
                    return [Completion('[{}]')], rules
                else:
                    return [Completion('[]')], rules
            else:
                return [Completion(json.dumps(rules))], rules
        else:
            token_after_colon = payload_tokens[index_colon + 1]
            last_payload_token = payload_tokens[-1]
            _logger.debug(f'The token after colon is: {token_after_colon}, last payload_token is: {last_payload_token}')
            if token_after_colon.ttype is BracketLeft:
                if token_after_colon is last_payload_token or (last_payload_token.ttype is Comma):
                    if isinstance(rules, list) and rules and isinstance(rules[0], dict):
                        return [Completion('{}')], rules
            elif token_after_colon.ttype in String and token_after_colon is last_payload_token:
                if isinstance(rules, dict) and '__one_of' in rules:
                    return [Completion(c) for c in rules['__one_of'] if isinstance(c, str)], rules
                elif isinstance(rules, list):
                    return [Completion(c) for c in rules if isinstance(c, str)], rules
                elif isinstance(rules, str):
                    return [Completion(rules)], rules
            elif token_after_colon.ttype is Name and token_after_colon is last_payload_token:
                if isinstance(rules, dict) and '__one_of' in rules:
                    return [Completion(json.dumps(c)) for c in rules['__one_of'] if c in (True, False, None)], rules
                elif isinstance(rules, list):
                    return [Completion(json.dumps(c)) for c in rules if c in (True, False, None)], rules

        return [], {}  # catch all

    def _find_rules_for_method_and_url_path(self, method: str, path_tokens: List[PeekToken]):
        token_stream = [t.value for t in path_tokens if t.ttype is PathPart]
        try:
            api_spec = next(matchable_specs(method, token_stream, self.specs, required_field='data_autocomplete_rules'))
            _logger.debug(f'Found API spec for {method!r} {path_tokens}')
        except StopIteration:
            _logger.debug(f'No matching API spec found for {method!r} {path_tokens}')
            return None

        rules = self._maybe_process_rules([], api_spec.get('data_autocomplete_rules', None))
        _logger.debug('No rules found from spec' if rules is None else f'Found rules from spec: {rules}')
        return rules

    def _resolve_rules_for_keys(self, rules, payload_keys, unwrap_value_for_last_key=True):
        rules = self._do_resolve_rules_for_keys(
            rules, payload_keys, unwrap_value_for_last_key=unwrap_value_for_last_key
        )

        # If the first key lookup did not get anything, try with the GLOBAL space
        if rules is None and payload_keys[0] in self.specs['GLOBAL']:
            _logger.debug('Retry key with GLOBAL')
            rules = self._do_resolve_rules_for_keys(
                self.specs['GLOBAL'], payload_keys, unwrap_value_for_last_key=unwrap_value_for_last_key
            )
        return rules

    def _do_resolve_rules_for_keys(self, rules, payload_keys, unwrap_value_for_last_key=True):
        history = [rules]
        for i, k in enumerate(payload_keys):
            if k not in rules and '*' in rules:
                _logger.debug(f'Matching * for key: {k!r}')
                rules = rules['*']
            elif k not in rules and '{field}' in rules:
                _logger.debug(f'Matching {{field}} for key: {k!r}')
                rules = rules['{field}']
            else:
                rules = rules.get(k, None)

            if i != len(payload_keys) - 1 or unwrap_value_for_last_key:
                rules = self._maybe_process_rules(history, rules)
            else:
                rules = self._maybe_process_rules(history, rules, unwrap_value=False)
            if k == 'query' and rules == {}:
                _logger.debug('Special handling for empty "query" field')
                rules = self.specs['GLOBAL']['query']
            elif k == 'script' and (rules == {} or rules is None):
                _logger.debug(f'Special handling for script field: {rules!r}')
                rules = self.specs['GLOBAL']['script']

            _logger.debug(f'Rules for key {k!r} is: {rules}')
            history.append(rules)
            if rules is None:
                break
        return rules

    def _maybe_process_rules(self, history, rules, unwrap_value=True):
        if unwrap_value:
            processors = [
                self._maybe_resolve_scope_link,
                self._maybe_unwrap_for_dict,
                self._maybe_lift_one_of,
            ]
        else:
            processors = [
                self._maybe_resolve_scope_link,
                self._maybe_lift_one_of,
            ]
        while True:
            original_rules = rules
            for p in processors:
                rules = p(history, rules)
            if rules == original_rules:
                break
        return rules

    def _maybe_resolve_scope_link(self, history, rules):
        if isinstance(rules, dict) and '__scope_link' in rules:
            rules = dict(rules)  # avoid mutating original value
            _logger.debug(f'Found scope link: {rules!r}')
            scope_link = rules.pop('__scope_link')
            if scope_link.startswith('.'):
                _logger.debug(f'Found relative scope link: {scope_link}')
                if len(history) >= 2:
                    rules = history[-2]
                    if scope_link[1:] != '':
                        for scope_link_key in scope_link[1:].split('.'):
                            rules = rules[scope_link_key]
                else:
                    _logger.warning(f'History is not long enough to support relative scope link: {len(history)}')

            elif '.' in scope_link:
                scope = self.specs
                for scope_link_key in scope_link.split('.'):
                    if scope_link_key in scope:
                        scope = scope[scope_link_key]
                if id(scope) != id(self.specs):
                    rules.update(scope['data_autocomplete_rules'] if 'data_autocomplete_rules' in scope else scope)
            else:
                scope = self.specs.get(scope_link, None)
                if isinstance(scope, dict):
                    rules.update(scope['data_autocomplete_rules'] if 'data_autocomplete_rules' in scope else scope)
        return rules

    def _maybe_lift_one_of(self, history, rules):
        if isinstance(rules, dict) and '__one_of' in rules:
            rules = dict(rules)
            one_of = rules.pop('__one_of')
            if isinstance(one_of[0], dict):
                for e in one_of:
                    rules.update(e)
            else:
                rules['__one_of'] = one_of
        return rules

    def _maybe_unwrap_for_dict(self, history, rules):
        """
        If the rules is an list of dict, return the first dict
        """
        if isinstance(rules, dict):
            return rules
        elif isinstance(rules, list) and len(rules) > 0:
            if isinstance(rules[0], dict):
                return rules[0]
            else:
                return None
        else:
            return None

    def _build_specs(self):
        specs = {}
        if not self.app.batch_mode and self.app.config.as_bool('load_api_specs'):
            _logger.info(f'Build API specs from: {self.kibana_dir}')
            try:
                specs.update(load_json_specs(self.kibana_dir))
            except Exception:
                _logger.exception('Error loading JSON specs')
            specs = _merge_specs(specs, self._build_js_specs())
        return specs

    def _build_js_specs(self):
        specs = {}
        if self.app.config.as_bool('build_extended_api_specs'):
            _logger.info(f'Build extended API specs from: {self.kibana_dir}')
            try:
                specs = build_js_specs(self.kibana_dir, self.app.config.as_bool('cache_extended_api_specs'))
            except Exception:
                _logger.exception('Error building JS specs')
        return specs


def _merge_specs(basic_specs, extended_specs):
    specs = dict(basic_specs)
    for k, v in extended_specs.items():
        if k not in specs:
            specs[k] = v
        else:
            for kk, vv in v.items():
                if kk in specs[k]:
                    _logger.debug(f'Duplicated key found: {k}.{kk}')
                    if isinstance(specs[k][kk], dict) and isinstance(vv, dict):
                        specs[k][kk].update(vv)
                else:
                    specs[k][kk] = vv
    return specs


def matchable_specs(method: str, ts: List[str], specs: Dict, required_field='url_params') -> Dict:
    """
    Find the matchable spec for the given HTTP method and input path.
    """
    for api_name, api_spec in specs.items():
        if 'methods' not in api_spec:
            continue
        if method not in api_spec['methods']:
            continue
        if not api_spec.get(required_field, None):
            continue
        matched = False
        for pattern in api_spec['patterns']:
            ps = [p for p in pattern.split('/') if p]
            if len(ts) != len(ps):
                continue
            if not can_match(ts, ps):
                continue
            matched = True
            break
        if matched:
            yield api_spec


def can_match(ts, ps):
    """
    Test whether the input path (ts) can match the candidate path (ps).
    The rule is basically a placeholder can match any string other than
    the ones leading with underscore.
    """
    for t, p in zip(ts, ps):
        if t != p:
            if t.startswith('_'):
                return False
            if not (p.startswith('{') and p.endswith('}')):
                return False
    return True



import json
import logging
import os
import ssl
import sys
import urllib
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from queue import Queue
from threading import Thread
from typing import Any, Optional

from peek.connection import EsClient, RefreshingEsClient
from peek.errors import PeekError

_logger = logging.getLogger(__name__)


class _SamlExchange:
    callback_path = None


class CallbackHTTPRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        pass

    def do_POST(self):
        content_length = int(self.headers['Content-Length'])
        body = self.rfile.read(content_length)
        _logger.debug(f'Read body: {body}')
        _SamlExchange.callback_path.put(body)
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b'Callback received, you can now close the browser tab.')

    def log_message(self, fmt: str, *args: Any) -> None:
        _logger.info("%s - - [%s] %s\n" % (self.address_string(), self.log_date_time_string(), fmt % args))


def saml_authenticate(es_client: EsClient, realm: str, callback_port: str, callback_ssl: bool, name: Optional[str]):
    _logger.info(f'SAML authenticate for realm {realm!r} and callback port {callback_port!r}')
    prepare_response = _saml_prepare(es_client, realm)
    httpd = _saml_start_http_server(callback_port, callback_ssl)
    print('Please use browser to complete authentication against the idP')
    webbrowser.open(prepare_response['redirect'])
    callback_path = _SamlExchange.callback_path.get()
    if isinstance(callback_path, bytes):
        callback_path = callback_path.decode('utf-8')
    try:
        httpd.shutdown()
        query = urllib.parse.parse_qs(callback_path)
        if 'SAMLResponse' not in query or len(query['SAMLResponse']) != 1:
            raise PeekError(f'Invalid saml callback response: {callback_path!r}')
        content = query['SAMLResponse'][0]
        auth_response = _saml_do_authenticate(es_client, realm, prepare_response['id'], content)
        return RefreshingEsClient(
            es_client,
            auth_response['username'],
            auth_response['access_token'],
            auth_response['refresh_token'],
            auth_response['expires_in'],
            name=name,
        )
    finally:
        _SamlExchange.callback_path = None


def _saml_prepare(es_client, realm: str):
    return es_client.perform_request(
        'POST',
        '/_security/saml/prepare',
        json.dumps(
            {
                'realm': realm,
            }
        ),
        deserialize_it=True,
    ).body


def _saml_do_authenticate(es_client, realm: str, _id: str, content: str):
    response = es_client.perform_request(
        'POST',
        '/_security/saml/authenticate',
        json.dumps(
            {
                'realm': realm,
                'ids': [_id],
                'content': content,
            }
        ),
        deserialize_it=True,
    ).body
    return response


def _saml_start_http_server(callback_port, callback_ssl):
    from peek import __file__ as package_root

    package_root = os.path.dirname(package_root)
    _SamlExchange.callback_path = Queue(maxsize=1)
    httpd = HTTPServer(('localhost', int(callback_port)), CallbackHTTPRequestHandler)
    if callback_ssl:
        keyfile = os.path.join(package_root, 'certs', 'key.pem')
        certfile = os.path.join(package_root, 'certs', 'cert.pem')

        if sys.version_info < (3, 12):
            httpd.socket = ssl.wrap_socket(httpd.socket, keyfile=keyfile, certfile=certfile, server_side=True)
        else:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)
            httpd.socket = ssl_context.wrap_socket(httpd.socket, server_side=True)

    t = Thread(target=httpd.serve_forever, daemon=True)
    t.start()
    return httpd


class SamlAuthenticateFunc:
    def __call__(self, app, **options):
        realm = options.get('realm', 'saml1')
        conn = options.get('conn', None)
        saml_es_client = saml_authenticate(
            app.es_client_manager.current if conn is None else app.es_client_manager.get_client(conn),
            realm,
            options.get('callback_port', '5601'),
            options.get('callback_ssl', True),
            name=options.get('name', None),
        )
        app.es_client_manager.add(saml_es_client)
        return json.dumps({'username': saml_es_client.username, 'realm': 'realm'})

    @property
    def options(self):
        return {'realm': 'saml1', 'callback_port': '5601', 'callback_ssl': True, 'name': None, 'conn': None}

    @property
    def description(self):
        return 'Start SAML authentication flow'



import os


class PeekError(Exception):
    pass


class PeekSyntaxError(PeekError):
    def __init__(self, text: str, error_token, title=None, message=None):
        self.text = text
        self.error_token = error_token
        self.title = title or 'Syntax error'
        self.message = message

    def __str__(self):
        text_before_error = self.text[: self.error_token.index]
        last_linesep = text_before_error.rfind(os.linesep)
        if last_linesep == -1:
            line = text_before_error
            line_index = 0
        else:
            line = text_before_error[last_linesep + len(os.linesep) :]
            line_index = text_before_error.count(os.linesep)
        col_index = len(line)

        text_since_error = self.text[self.error_token.index :]
        next_linesep = text_since_error.find(os.linesep)
        if next_linesep == -1:
            line += text_since_error
        else:
            line += text_since_error[:next_linesep]

        return (
            f'{self.title} at Line {line_index + 1}, Column {col_index + 1}:\n'
            f'{line}\n'
            f'{" " * col_index}'
            f'{"^" * len(self.error_token.value)}'
            f'{os.linesep + self.message if self.message else ""}'
        )


class InvalidEsApiCall(PeekError):
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return f'Invalid ES API call. Expect format of "METHOD URL [PAYLOAD]", got [{self.text}]'


class InvalidHttpMethod(PeekError):
    def __init__(self, method: str):
        self.method = method

    def __str__(self):
        return f'Invalid HTTP method: [{self.method}]'



"""Main module."""
import json
import logging
import logging.handlers
import sys
from datetime import datetime
from typing import Iterable

from prompt_toolkit import PromptSession, prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.layout.processors import HighlightMatchingBracketProcessor
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments_cls

from peek.capture import FileCapture, NoOpCapture
from peek.common import AUTO_SAVE_NAME, NONE_NS
from peek.completer import PeekCompleter
from peek.completions import monkey_patch_completion_state
from peek.config import config_location, get_config
from peek.connection import DelegatingListener, EsClientManager, connect
from peek.display import Display
from peek.errors import PeekError, PeekSyntaxError
from peek.history import SqLiteHistory
from peek.key_bindings import key_bindings
from peek.lexers import Heading, PeekLexer, PeekStyle, TipsMinor
from peek.parser import PeekParser
from peek.vm import PeekVM

_logger = logging.getLogger(__name__)


class PeekApp:
    def __init__(
        self, batch_mode=False, config_file: str = None, extra_config_options: Iterable[str] = None, cli_ns=NONE_NS
    ):
        self._should_exit = False
        self._preserved_text = ''
        self.capture = NoOpCapture()
        self.batch_mode = batch_mode
        self.config = get_config(config_file, extra_config_options)
        self.cli_ns = cli_ns
        self._init_logging()
        self.history = SqLiteHistory(self.config.as_int('history_max'))
        self.completer = PeekCompleter(self)
        self.display = Display(self)
        self.parser = PeekParser()
        self.vm = self._init_vm()
        self.prompt = self._init_prompt()
        monkey_patch_completion_state()
        # TODO: better name for signal payload json reformat
        self.is_pretty = True
        self._init_es_client_manager()
        self.ecm_backup_data = self.es_client_manager.to_dict()
        self._on_startup()

    def run(self):
        try:
            while not self._should_exit:
                try:
                    message = self._get_message()
                    if self.capture.file() is not None:
                        print(''.join(v[1] for v in message), file=self.capture.file())
                    text: str = self.prompt.prompt(
                        message=message,
                        default=self._get_default_text(),
                    )
                    _logger.debug(f'input: {text!r}')
                    if self._should_exit:
                        raise EOFError
                    if text.strip() == '':
                        continue
                    self.process_input(text)
                except KeyboardInterrupt:
                    continue
                except EOFError:
                    break
        finally:
            self.on_exit()

    def process_input(self, text, echo=False):
        try:
            nodes = self.parser.parse(text)
        except PeekSyntaxError as e:
            self.display.error(e)
            return

        for node in nodes:
            try:
                if echo:
                    self.display.info(str(node))
                self.execute_node(node)
            except PeekError as e:
                self.display.error(e)
            except Exception as e:
                self.display.error(e)
                _logger.exception('Error on node execution')

    def execute_node(self, node):
        self.vm.execute_node(node)

    def signal_exit(self):
        self._should_exit = True

    def start_capture(self, f=None):
        if not isinstance(self.capture, NoOpCapture):
            raise PeekError(f'Cannot capture when one is currently running: {self.capture.status()}')

        if f is None:
            f = f'{datetime.now().strftime("%Y%m%d%H%M%S")}.es'
        self.capture = FileCapture(f)
        return self.capture.status()

    def stop_capture(self):
        if not isinstance(self.capture, NoOpCapture):
            self.capture.stop()
            self.capture = NoOpCapture()
            return 'Capture stopped'
        else:
            return 'No capture is running'

    @property
    def preserved_text(self):
        return self._preserved_text

    @preserved_text.setter
    def preserved_text(self, value: str):
        self._preserved_text = value

    def input(self, message='', is_secret=False):
        return prompt(message=message, is_password=is_secret)

    def reset(self):
        self.completer.init_api_completer()
        self.vm = self._init_vm()
        self._repopulate_clients(EsClientManager.from_dict(self, self.ecm_backup_data))
        self._on_startup()

    def _get_message(self):
        idx = self.es_client_manager.index_current
        if idx is None:
            return FormattedText(
                [
                    (PeekStyle.styles[Heading], '>>>'),
                    (PeekStyle.styles[TipsMinor], ' No Connection\n'),
                ]
            )
        else:
            info_line = f' [{idx}] {self.es_client_manager.current}'
            if len(info_line) > 100:
                info_line = info_line[:97] + '...'
            return FormattedText(
                [
                    (PeekStyle.styles[Heading], '>>>'),
                    (PeekStyle.styles[TipsMinor], info_line + '\n'),
                ]
            )

    def _get_default_text(self):
        text = self.preserved_text
        if text:
            self.preserved_text = ''
        return text

    def _init_logging(self):
        log_file = self.config['log_file']
        log_level = self.config['log_level'].upper()
        if log_level == 'NONE':
            handler = logging.NullHandler()
        elif log_file == 'stderr':
            handler = logging.StreamHandler(sys.stderr)
        elif log_file == 'stdout':
            handler = logging.StreamHandler(sys.stdout)
        else:
            handler = logging.handlers.RotatingFileHandler(
                config_location() + log_file, maxBytes=10 * 1024 * 1024, backupCount=5
            )

        log_level = getattr(logging, log_level, logging.WARNING)

        formatter = logging.Formatter(
            '%(asctime)s (%(process)d/%(threadName)s) ' '%(name)s %(levelname)s - %(message)s'
        )
        handler.setFormatter(formatter)

        root_logger = logging.getLogger(__package__)
        root_logger.addHandler(handler)
        root_logger.setLevel(log_level)

    def _init_prompt(self):
        if self.batch_mode:
            return None
        else:
            try:
                from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard

                clipboard = PyperclipClipboard()
            except ImportError:
                _logger.info('pyperclip not available, disable clipboard integration')
                clipboard = None

            return PromptSession(
                style=style_from_pygments_cls(PeekStyle),
                lexer=PygmentsLexer(PeekLexer),
                auto_suggest=AutoSuggestFromHistory(),
                completer=self.completer,
                history=self.history,
                multiline=True,
                key_bindings=key_bindings(self),
                enable_open_in_editor=True,
                enable_system_prompt=False,
                enable_suspend=True,
                search_ignore_case=True,
                clipboard=clipboard,
                mouse_support=Condition(self._should_support_mouse),
                swap_light_and_dark_colors=self.config.as_bool('swap_colour'),
                input_processors=[
                    HighlightMatchingBracketProcessor(chars="[](){}"),
                ],
            )

    def _init_es_client_manager(self):
        if self.cli_ns.zero_connection:
            self._repopulate_clients(EsClientManager())
            return

        options = {}
        keys = [
            'name',
            'hosts',
            'cloud_id',
            'username',
            'password',
            'api_key',
            'token',
            'use_ssl',
            'verify_certs',
            'assert_hostname',
            'assert_fingerprint',
            'ca_certs',
            'client_cert',
            'client_key',
            'force_prompt',
            'no_prompt',
        ]
        for k in keys:
            v = getattr(self.cli_ns, k, None)
            if v is not None:
                options[k] = v

        if self._should_auto_load_session(options):
            _logger.info('Auto-loading connection state')
            data = self.history.load_session(AUTO_SAVE_NAME)
            if data is not None:
                self._repopulate_clients(EsClientManager.from_dict(self, json.loads(data)))
                return

        es_client_manager = EsClientManager()
        es_client_manager.add(connect(self, **options))
        self._repopulate_clients(es_client_manager)

    def _repopulate_clients(self, m: EsClientManager):
        """
        Detour to populate the clients due to dependency, e.g. on_add callback gets a reference of the App
        """
        self.es_client_manager = EsClientManager(listeners=[DelegatingListener(on_add=self._on_client_add)])
        for es_client in m.clients():
            self.es_client_manager.add(es_client)
        idx_current = m.index_current
        if idx_current is not None and idx_current < len(self.es_client_manager.clients()):
            self.es_client_manager.set_current(idx_current)

    def _init_vm(self):
        return PeekVM(self)

    def on_exit(self):
        if not self.batch_mode and self.config.as_bool('auto_save_session'):
            _logger.info('Auto-saving connection state')
            data = self.es_client_manager.to_dict()
            self.history.save_session(AUTO_SAVE_NAME, json.dumps(data))

    def _should_auto_load_session(self, options):
        """
        Auto load last auto-saved session only when:
        * configuration says so
        * running in interactive mode
        * No explicit connection parameters are specified (include those specified in connection section of rc file)
        """
        return (
            self.config.as_bool('auto_load_session')
            and not self.batch_mode
            and not options
            and self.config.get('connection') is None
        )

    def _should_support_mouse(self) -> bool:
        return self.config.as_bool('mouse_support')

    def _on_client_add(self, es_client_manager: EsClientManager):
        on_connection_add = self.config.get('on_connection_add')
        if on_connection_add is not None:
            self.process_input(on_connection_add)

    def _on_startup(self):
        on_startup = self.config.get('on_startup')
        if on_startup is not None:
            self.process_input(on_startup)



import ast
import json
import logging
from abc import ABCMeta
from typing import List, Tuple

from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.document import Document
from pygments.token import Name, String

from peek.common import PeekToken
from peek.es_api_spec.schema import Schema
from peek.lexers import EOF, Assign, BracketLeft, Colon, Comma, CurlyLeft, CurlyRight, DictKey, PathPart, Slash
from peek.parser import ParserEvent, ParserEventType

_logger = logging.getLogger(__name__)


class ESApiCompleter(metaclass=ABCMeta):
    def complete_url_path(
        self, document: Document, complete_event: CompleteEvent, method: str, path_tokens: List[PeekToken]
    ) -> List[Completion]:
        return []

    def complete_query_param_name(
        self, document: Document, complete_event: CompleteEvent, method: str, path_tokens: List[PeekToken]
    ) -> List[Completion]:
        return []

    def complete_query_param_value(
        self, document: Document, complete_event: CompleteEvent, method: str, path_tokens: List[PeekToken]
    ) -> List[Completion]:
        return []

    def complete_payload(
        self,
        document: Document,
        complete_event: CompleteEvent,
        method: str,
        path_tokens: List[PeekToken],
        payload_tokens: List[PeekToken],
        payload_events: List[ParserEvent],
    ) -> Tuple[List[Completion], dict]:
        return [], {}

    def complete_payload_value(
        self,
        document: Document,
        complete_event: CompleteEvent,
        method: str,
        path_tokens: List[PeekToken],
        payload_tokens: List[PeekToken],
        payload_events: List[ParserEvent],
    ) -> Tuple[List[Completion], dict]:
        return [], {}


class NoopESApiCompleter(ESApiCompleter):
    pass


class SchemaESApiCompleter(ESApiCompleter):
    def __init__(self, schema_filepath):
        with open(schema_filepath) as ins:
            self._schema = Schema(json.load(ins))

    def complete_url_path(self, document, complete_event, method, path_tokens):
        cursor_token = path_tokens[-1]
        _logger.debug(f'Completing URL path: {cursor_token}')
        token_stream = [t.value for t in path_tokens if t.ttype is not Slash]
        if cursor_token.ttype is PathPart:
            token_stream.pop()
        return [Completion(c) for c in self._schema.candidate_urls(method, token_stream)]

    def complete_query_param_name(self, document, complete_event, method, path_tokens):
        _logger.debug(f'Completing URL query param name: {path_tokens[-1]}')
        token_stream = [t.value for t in path_tokens if t.ttype is PathPart]
        return [Completion(c) for c in self._schema.candidate_query_param_names(method, token_stream)]

    def complete_query_param_value(self, document, complete_event, method, path_tokens):
        _logger.debug(f'Completing URL query param value: {path_tokens[-1]}')
        param_name_token = path_tokens[-2] if path_tokens[-1].ttype is Assign else path_tokens[-3]
        ts = [t.value for t in path_tokens if t.ttype is PathPart]
        return [Completion(c) for c in self._schema.candidate_query_param_values(method, ts, param_name_token.value)]

    def complete_payload(self, document, complete_event, method, path_tokens, payload_tokens, payload_events):
        _logger.debug(f'Completing for API payload: {method!r} {path_tokens!r} {payload_tokens!r}')
        # TODO: refactor with parser state tracker
        payload_keys = []
        curly_level = 0
        for t in payload_tokens[:-1]:
            if t.ttype is CurlyLeft:
                curly_level += 1
            elif t.ttype is CurlyRight:
                curly_level -= 1
                payload_keys.pop()
            elif t.ttype is DictKey:
                if len(payload_keys) == curly_level - 1:
                    payload_keys.append(ast.literal_eval(t.value))
                elif len(payload_keys) == curly_level:
                    payload_keys[-1] = ast.literal_eval(t.value)
                else:
                    raise ValueError(f'Error when counting curly level {curly_level} and keys {payload_keys}')

        _logger.debug(f'Payload status: level: {curly_level}, keys: {payload_keys}')
        if curly_level == 0:  # not even in the first curly bracket, no completion
            return [], {}

        # Remove the payload key that is at the same level
        if curly_level == len(payload_keys):
            payload_keys.pop()

        ts = [t.value for t in path_tokens if t.ttype is PathPart]
        key_to_values = self._schema.candidate_sub_key_values(method, ts, payload_keys)
        return [Completion(c) for c in sorted(key_to_values.keys())], key_to_values

    def complete_payload_value(self, document, complete_event, method, path_tokens, payload_tokens, payload_events):
        _logger.debug(f'Completing for API payload value: {method!r} {path_tokens!r} {payload_tokens!r}')
        completions = self._do_complete_payload_value(
            document, complete_event, method, path_tokens, payload_tokens, payload_events
        )
        return [Completion(c) for c in sorted(set(completions))], {}

    def _do_complete_payload_value(self, document, complete_event, method, path_tokens, payload_tokens, payload_events):
        unpaired_dict_key_tokens = []
        for payload_event in payload_events:
            if payload_event.type is ParserEventType.BEFORE_DICT_KEY_EXPR:
                _logger.debug(f'No completion is possible for payload with dict key expr: {payload_event.token}')
                return []
            elif payload_event.type is ParserEventType.DICT_KEY:
                unpaired_dict_key_tokens.append(payload_event.token)
            elif payload_event.type is ParserEventType.AFTER_DICT_VALUE and payload_event.token.ttype is not EOF:
                unpaired_dict_key_tokens.pop()

        if not unpaired_dict_key_tokens:  # should not happen
            _logger.warning('No unpaired dict key tokens are found')
            return []

        payload_keys = [ast.literal_eval(t.value) for t in unpaired_dict_key_tokens]
        _logger.debug(f'Payload keys are: {payload_keys}')

        # Find last Colon position
        for index_colon in range(len(payload_tokens) - 1, -1, -1):
            if payload_tokens[index_colon].ttype is Colon:
                break
        else:
            _logger.warning(f'Should not happen - Colon not found in payload: {payload_tokens}')
            return []

        ts = [t.value for t in path_tokens if t.ttype is PathPart]
        if index_colon == len(payload_tokens) - 1:  # Colon is the last token
            _logger.debug('Colon is the last token')
            # The simpler case when value position has nothing yet
            return [json.dumps(v) for v in self._schema.candidate_values(method, ts, payload_keys)]
        else:
            token_after_colon = payload_tokens[index_colon + 1]
            last_payload_token = payload_tokens[-1]
            _logger.debug(f'The token after colon is: {token_after_colon}, last payload_token is: {last_payload_token}')
            if token_after_colon.ttype is BracketLeft:
                if last_payload_token.ttype is BracketLeft or last_payload_token.ttype is Comma:
                    # no question asked, let's complete
                    values = self._schema.candidate_values(method, ts, payload_keys, inside_array=True)
                    return [json.dumps(v) for v in values]
                elif last_payload_token.ttype in String:
                    try:
                        ast.literal_eval(last_payload_token.value)
                    except SyntaxError:
                        # String is not complete, i.e. we are completing the string under cursor
                        values = self._schema.candidate_values(method, ts, payload_keys, inside_array=True)
                        return [v for v in values if isinstance(v, str)]

            elif token_after_colon.ttype in String and token_after_colon is last_payload_token:
                return [v for v in self._schema.candidate_values(method, ts, payload_keys) if isinstance(v, str)]
            elif token_after_colon.ttype is Name and token_after_colon is last_payload_token:
                return [
                    json.dumps(v)
                    for v in self._schema.candidate_values(method, ts, payload_keys)
                    if v in (True, False, None)
                ]

        return []  # catch all



from abc import ABCMeta


class Capture(metaclass=ABCMeta):
    def stop(self):
        pass

    def status(self):
        pass

    def file(self):
        pass


class NoOpCapture(Capture):
    def status(self):
        return 'No capture is running'


class FileCapture(Capture):
    def __init__(self, f):
        self.f = f
        self.outs = open(self.f, 'w')

    def stop(self):
        self.outs.close()

    def status(self):
        return f'Capture with file: {self.f}'

    def file(self):
        return self.outs



import itertools
import logging
import os
from typing import Iterable, List, Optional

from prompt_toolkit.completion import CompleteEvent, Completer, Completion, FuzzyCompleter, PathCompleter, WordCompleter
from prompt_toolkit.contrib.completers import SystemCompleter
from prompt_toolkit.document import Document
from pygments.token import Error, Literal, Name, String

from peek.common import HTTP_METHODS, PeekToken
from peek.completions import PayloadKeyCompletion
from peek.config import config_location
from peek.lexers import (
    EOF,
    Ampersand,
    DictKey,
    FuncName,
    HttpMethod,
    ParamName,
    PathPart,
    PeekLexer,
    QuestionMark,
    ShellOut,
    Slash,
    UrlPathLexer,
)
from peek.parser import ParserEvent, ParserEventType, PeekParser

_logger = logging.getLogger(__name__)

_HTTP_METHOD_COMPLETER = WordCompleter([m.upper() for m in HTTP_METHODS], ignore_case=True)

_ES_API_CALL_OPTION_NAME_COMPLETER = WordCompleter(
    [w + '=' for w in sorted(['conn', 'runas', 'headers', 'xoid', 'quiet', 'out', 'pipe'])]
)

_PATH_COMPLETER = PathCompleter(expanduser=True)
_SYSTEM_COMPLETER = SystemCompleter()

_DICT_EVENT_TYPES = (
    ParserEventType.DICT_KEY,
    ParserEventType.BEFORE_DICT_KEY_EXPR,
    ParserEventType.AFTER_DICT_KEY_EXPR,
    ParserEventType.BEFORE_DICT_VALUE,
    ParserEventType.AFTER_DICT_VALUE,
)


class ParserStateTracker:
    def __init__(self, text: str):
        self.text = text
        self._events: List[ParserEvent] = []
        self._tokens: List[PeekToken] = []
        self._payload_events: List[ParserEvent] = []

    def __call__(self, event: ParserEvent):
        if event.type is ParserEventType.AFTER_TOKEN:
            self._tokens.append(event.token)
        elif event.type in _DICT_EVENT_TYPES:
            if self.last_event.type is ParserEventType.BEFORE_ES_PAYLOAD_INLINE:
                self._payload_events.append(event)
            else:
                _logger.debug(f'Ignore dict parsing events for non-payload: {self.last_event}')
        else:
            self._events.append(event)
            if event is ParserEventType.BEFORE_ES_PAYLOAD_INLINE:
                self._payload_events = []

    @property
    def events(self):
        return self._events

    @property
    def tokens(self):
        return self._tokens

    @property
    def payload_events(self):
        return self._payload_events

    @property
    def stmt_token(self) -> Optional[PeekToken]:
        if not self._events:
            return None
        return self._events[0].token

    @property
    def is_completion_possible(self):
        if self.stmt_token is None or not self._tokens:
            return False
        # The last non-white char must be parsed successfully for completion to be available
        last_token = self.last_token
        if last_token is None:
            return False
        if self.text[last_token.index :].rstrip() == last_token.value:
            return True
        else:
            return False

    @property
    def last_token(self) -> Optional[PeekToken]:
        if not self._tokens:
            return None
        return self._tokens[-1]

    @property
    def last_event(self) -> Optional[ParserEvent]:
        if not self._events:
            return None
        return self._events[-1]

    @property
    def last_payload_event(self) -> Optional[ParserEvent]:
        if not self._payload_events:
            return None
        return self._payload_events[-1]

    @property
    def is_cursor_on_whitespace(self):
        last_token = self.last_token
        if last_token is None:
            return True
        return len(self.text) > (last_token.index + len(last_token.value))

    @property
    def is_within_payload_value(self):
        last_payload_event = self.last_payload_event
        if last_payload_event is None:
            return False
        return last_payload_event.type is ParserEventType.BEFORE_DICT_VALUE or (
            last_payload_event.type is ParserEventType.AFTER_DICT_VALUE and last_payload_event.token.ttype is EOF
        )

    @property
    def has_newline_after_last_token(self):
        last_token = self.tokens[-1]
        return '\n' in self.text[last_token.index :]


class PeekCompleter(Completer):
    def __init__(self, app):
        self.app = app
        self.lexer = PeekLexer()
        self.url_path_lexer = UrlPathLexer()
        self.api_completer = self.init_api_completer()

    def init_api_completer(self):
        from peek import __file__ as package_root

        package_root = os.path.dirname(package_root)
        if self.app.config.as_bool('prefer_elasticsearch_specification'):
            _logger.info('Use elasticsearch-specification schema for autocompletion')
            from peek.es_api_spec.api_completer import SchemaESApiCompleter

            schema_filepath = os.path.join(config_location(), 'schema.json')
            if not os.path.exists(schema_filepath):
                schema_filepath = os.path.join(package_root, 'specs', 'schema.json')
            if os.path.exists(schema_filepath):
                return SchemaESApiCompleter(schema_filepath)
            else:
                from peek.es_api_spec.api_completer import NoopESApiCompleter

                return NoopESApiCompleter()
        else:
            kibana_dir = self.app.config['kibana_dir']
            if not kibana_dir:
                config_dir = config_location()
                if os.path.exists(config_dir):
                    kibana_dirs = [
                        os.path.join(config_dir, d) for d in os.listdir(config_dir) if d.startswith('kibana-')
                    ]
                    if kibana_dirs:
                        kibana_dir = kibana_dirs[0]
            if not kibana_dir:
                kibana_dir = os.path.join(package_root, 'specs', 'kibana-7.8.1')
            _logger.info(f'Attempt to build Elasticsearch API specs from: {kibana_dir}')
            from peek.es_api_spec.kspec import KibanaSpecESApiCompleter

            return KibanaSpecESApiCompleter(self.app, kibana_dir)

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        _logger.debug(f'Document: {document}, Event: {complete_event}')

        text_before_cursor = document.text_before_cursor
        _logger.debug(f'Text before cursor: {text_before_cursor!r}')
        if text_before_cursor.strip() == '':
            return []

        state_tracker = ParserStateTracker(text_before_cursor)
        try:
            PeekParser((state_tracker,)).parse(
                text_before_cursor, fail_fast_on_error_token=True, last_stmt_only=True, log_level='WARNING'
            )
        except Exception:
            pass

        if not state_tracker.is_completion_possible:
            _logger.debug('No completion is available according to state_tracker')
            return []

        stmt_token = state_tracker.stmt_token
        if stmt_token.ttype is ShellOut:
            _logger.debug('Completing for shell out')
            return _SYSTEM_COMPLETER.get_completions(
                Document(text_before_cursor[stmt_token.index + 1 :]), complete_event
            )

        if state_tracker.is_cursor_on_whitespace:
            return self._get_completions_for_whitespace(document, complete_event, state_tracker)

        else:
            return self._get_completions_for_non_white(document, complete_event, state_tracker)

    def _get_completions_for_whitespace(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ) -> Iterable[Completion]:
        _logger.debug('Cursor is on a whitespace')
        stmt_token = state_tracker.stmt_token
        last_token = state_tracker.last_token
        last_event = state_tracker.last_event
        _logger.debug(f'Last Event: {last_event}, Last Token: {last_token}')
        has_newline_after_last_token = state_tracker.has_newline_after_last_token

        # Only option name/value completions are available when cursor is on whitespace char
        if stmt_token.ttype is HttpMethod:
            # HTTP option name/value completion is only possible if they are at the same line as the HTTP method
            if not has_newline_after_last_token and last_event.type in (
                ParserEventType.ES_URL,
                ParserEventType.AFTER_ES_URL_EXPR,
                ParserEventType.AFTER_ES_OPTION_VALUE,
            ):
                return _ES_API_CALL_OPTION_NAME_COMPLETER.get_completions(document, complete_event)
            elif not has_newline_after_last_token and last_event.type is ParserEventType.BEFORE_ES_OPTION_VALUE:
                return []  # TODO: ES option value
            elif last_event.type is ParserEventType.BEFORE_ES_PAYLOAD_INLINE:
                if state_tracker.last_payload_event.type is ParserEventType.BEFORE_DICT_VALUE:
                    return self._maybe_complete_payload_value(document, complete_event, state_tracker)

        elif stmt_token.ttype is FuncName and not has_newline_after_last_token:
            if last_event.type in (ParserEventType.FUNC_STMT, ParserEventType.AFTER_FUNC_ARGS):
                return self._maybe_complete_func_option_name(document, complete_event, state_tracker)
            elif last_event.type is ParserEventType.BEFORE_FUNC_OPTION_VALUE:
                return []  # TODO: Func option value
        else:
            return []  # TODO: Options of func expr

        return []

    def _get_completions_for_non_white(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ) -> Iterable[Completion]:
        _logger.debug('Cursor is on an non-white char')
        stmt_token = state_tracker.stmt_token
        last_token = state_tracker.last_token
        last_event = state_tracker.last_event
        _logger.debug(f'Last Event: {last_event}, Last Token: {last_token}')

        if len(state_tracker.tokens) == 1:
            _logger.debug(f'Completing for beginning of stmt: {stmt_token}')
            return itertools.chain(
                _HTTP_METHOD_COMPLETER.get_completions(document, complete_event),
                WordCompleter(['for', 'let']).get_completions(document, complete_event),
                WordCompleter(self.app.vm.functions.keys()).get_completions(document, complete_event),
            )

        elif stmt_token.ttype is HttpMethod:
            if last_event.type is ParserEventType.ES_URL:
                # ES_URL not ES_METHOD because we do not provide completion for empty string
                return self._maybe_complete_http_path(document, complete_event, state_tracker)
            elif last_event.type is ParserEventType.ES_OPTION_NAME:
                return _ES_API_CALL_OPTION_NAME_COMPLETER.get_completions(document, complete_event)
            elif last_event.type is ParserEventType.BEFORE_ES_OPTION_VALUE:
                return []  # TODO: ES option value
            elif last_event.type is ParserEventType.ES_PAYLOAD_FILE_AT:
                return _PATH_COMPLETER.get_completions(
                    Document(state_tracker.text[last_event.token.index + 1 :]), complete_event
                )
            elif last_event.type is ParserEventType.BEFORE_ES_PAYLOAD_INLINE:
                _logger.debug(f'Last payload event: {state_tracker.last_payload_event}')
                if last_token.ttype is DictKey:
                    return self._maybe_complete_payload(document, complete_event, state_tracker)
                elif state_tracker.is_within_payload_value:
                    return self._maybe_complete_payload_value(document, complete_event, state_tracker)

        elif stmt_token.ttype is FuncName:
            if self._function_option_name_can_appear(last_event, last_token):
                return self._maybe_complete_func_option_name(document, complete_event, state_tracker)
            elif last_event.type is ParserEventType.BEFORE_FUNC_OPTION_VALUE:
                return []  # TODO: Function option value
            elif stmt_token.value == 'run':
                return self._maybe_complete_special_for_run_func_file_path(document, complete_event, state_tracker)

        else:
            return []  # TODO: func expr

        return []

    def _function_option_name_can_appear(self, last_event: ParserEvent, last_token: PeekToken) -> bool:
        return (
            last_event.type is ParserEventType.AFTER_FUNC_ARGS and last_token.ttype in (Literal, Name)
        ) or last_event.type is ParserEventType.BEFORE_FUNC_SYMBOL_ARG

    def _maybe_complete_func_option_name(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ) -> Iterable[Completion]:
        stmt_token = state_tracker.stmt_token
        func = self.app.vm.functions.get(stmt_token.value)
        if func is None or not getattr(func, 'options', None):
            return []
        options = sorted([n if n.startswith('@') else (n + '=') for n in func.options.keys()])
        return WordCompleter(options, WORD=True).get_completions(document, complete_event)

    def _maybe_complete_special_for_run_func_file_path(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ):
        if len(state_tracker.tokens) == 2 and state_tracker.tokens[-1].ttype in String:
            last_token = state_tracker.last_token
            if last_token.ttype in (String.Single, String.Double):
                doc = Document(last_token.value[1:])
            else:
                doc = Document(last_token.value[3:])
            _logger.debug(f'Completing for file path for run: {doc}')
            return _PATH_COMPLETER.get_completions(doc, complete_event)
        else:
            return []

    def _maybe_complete_http_path(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ) -> Iterable[Completion]:
        tokens = state_tracker.tokens
        method_token, path_token = tokens[-2], tokens[-1]
        method = method_token.value.upper()
        cursor_position = document.cursor_position - path_token.index
        path = path_token.value[:cursor_position]
        _logger.debug(f'Completing HTTP API url: {path!r}')
        path_tokens = list(self.url_path_lexer.get_tokens_unprocessed(path))
        _logger.debug(f'URL Path Tokens: {path_tokens}')
        if not path_tokens:  # empty, should not happen
            return []

        cursor_token = path_tokens[-1]
        _logger.debug(f'Cursor Token: {cursor_token}')
        if cursor_token.ttype is Error:
            return []
        else:
            if cursor_token.ttype in (PathPart, Slash):
                complete_func = self.api_completer.complete_url_path
            elif cursor_token.ttype in (ParamName, QuestionMark, Ampersand):
                complete_func = self.api_completer.complete_query_param_name
            else:
                complete_func = self.api_completer.complete_query_param_value
            candidates = complete_func(document, complete_event, method, path_tokens)
            return FuzzyCompleter(ConstantCompleter(candidates)).get_completions(document, complete_event)

    def _maybe_complete_payload(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ) -> Iterable[Completion]:
        tokens = state_tracker.tokens
        _logger.debug(f'Completing for payload with tokens: {tokens}')
        path_event = state_tracker.events[1]
        if path_event.token.ttype is not Literal:  # No completion for non-literal URL
            return []
        method_token, path_token = tokens[0], tokens[1]
        path_tokens = list(self.url_path_lexer.get_tokens_unprocessed(path_token.value))
        last_event = state_tracker.last_event
        payload_tokens = tokens[tokens.index(last_event.token) :]
        candidates, rules = self.api_completer.complete_payload(
            document,
            complete_event,
            method_token.value.upper(),
            path_tokens,
            payload_tokens,
            state_tracker.payload_events,
        )
        if not candidates:
            return []
        constant_completer = ConstantCompleter(candidates)
        for c in FuzzyCompleter(constant_completer).get_completions(document, complete_event):
            yield PayloadKeyCompletion(
                c.text, rules[c.text], c.start_position, c.display, c.display_meta, c.style, c.selected_style
            )

    def _maybe_complete_payload_value(
        self, document: Document, complete_event: CompleteEvent, state_tracker: ParserStateTracker
    ) -> Iterable[Completion]:
        tokens = state_tracker.tokens
        _logger.debug(f'Completing for payload value with tokens: {tokens}')
        path_event = state_tracker.events[1]
        if path_event.token.ttype is not Literal:  # No completion for non-literal URL
            return []
        method_token, path_token = tokens[0], tokens[1]
        path_tokens = list(self.url_path_lexer.get_tokens_unprocessed(path_token.value))
        last_event = state_tracker.last_event
        candidates, rules = self.api_completer.complete_payload_value(
            document,
            complete_event,
            method_token.value.upper(),
            path_tokens,
            tokens[tokens.index(last_event.token) :],
            state_tracker.payload_events,
        )
        if not candidates:
            return []
        constant_completer = ConstantCompleter(candidates)
        return FuzzyCompleter(constant_completer).get_completions(document, complete_event)


class ConstantCompleter(Completer):
    def __init__(self, candidates):
        self.candidates = candidates

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        return self.candidates



import json
import logging
import sys
from json import JSONDecodeError
from typing import Any

import pygments
from prompt_toolkit import HTML, print_formatted_text
from prompt_toolkit.formatted_text import FormattedText, PygmentsTokens, merge_formatted_text, to_formatted_text
from prompt_toolkit.styles import (
    ConditionalStyleTransformation,
    SwapLightAndDarkStyleTransformation,
    style_from_pygments_cls,
)
from pygments.token import Token

from peek.lexers import Heading, PeekLexer, PeekStyle, TipsMinor

_logger = logging.getLogger(__name__)


class Display:
    def __init__(self, app):
        self.app = app
        self.payload_lexer = PeekLexer(stack=('value',))
        self.style = style_from_pygments_cls(PeekStyle)
        self.style_transformation = ConditionalStyleTransformation(
            SwapLightAndDarkStyleTransformation(), self.app.config.as_bool('swap_colour')
        )

    @property
    def pretty_print(self):
        return self.app.config.as_bool('pretty_print')

    def info(self, source, header_text=''):
        if source is None:
            return
        if not self.app.batch_mode:
            self._tee_print(
                FormattedText([(PeekStyle.styles[Heading], '=== '), (PeekStyle.styles[TipsMinor], header_text)]),
                plain_source=f'=== {header_text}',
            )
        if isinstance(source, FormattedText):
            self._tee_print(source)
        else:
            source, plain_source = self._try_jsonify(source)  # TODO: try more types
            self._tee_print(source, plain_source=plain_source)

    def error(self, source, header_text=''):
        if source is None:
            return
        if not self.app.batch_mode:
            self._tee_print(
                merge_formatted_text(
                    [
                        HTML('<ansired>--- </ansired>').formatted_text,
                        FormattedText([(PeekStyle.styles[TipsMinor], header_text)]),
                    ]
                ),
                plain_source=f'--- {header_text}',
            )
        if isinstance(source, FormattedText):
            self._tee_print(source)
        else:
            self._tee_print(source, plain_source=source)

    def warn(self, source):
        if source is None:
            return
        if isinstance(source, FormattedText):
            self._tee_print(source)
        else:
            self._tee_print(
                FormattedText(
                    [
                        ('#ffdf5d', f'WARNING: {source}'),
                    ]
                ),
                plain_source=f'WARNING: {source}',
            )

    def _try_jsonify(self, source):
        # If it is a string, first check whether it can be decoded as JSON
        if isinstance(source, str):
            try:
                source = json.loads(source)
            except JSONDecodeError:
                _logger.debug(f'Source string is not JSON: {source!r}')

        try:
            if not isinstance(source, str):
                source = json.dumps(source, cls=PeekEncoder, app=self.app, indent=2 if self.pretty_print else None)
            tokens = []
            for t in pygments.lex(source, lexer=self.payload_lexer):
                tokens.append(t)
                if t[0] is Token.Error:
                    _logger.debug(f'Source string is not valid payload type: {t!r}')
                    return source, source
            return PygmentsTokens(tokens), source
        except Exception as e:
            _logger.debug(f'Cannot render object as json: {source!r}, {e}')
            return source, source

    def _tee_print(self, source, plain_source=None):
        content = None
        if self.app.batch_mode and not sys.stdout.isatty():
            content = all_to_text(source) if plain_source is None else plain_source
            print(content, file=sys.stdout, end='')
        else:
            print_formatted_text(source, style=self.style, style_transformation=self.style_transformation)

        if self.app.capture.file() is not None:
            content = content or (all_to_text(source) if plain_source is None else plain_source)
            print(content, file=self.app.capture.file())


class PeekEncoder(json.JSONEncoder):
    def __init__(self, app=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.app = app
        assert self.app is not None, 'Parameter app must be provided'

    def default(self, o: Any) -> Any:
        if self.app is not None and callable(o):
            function_lookup = {v: k for k, v in self.app.vm.functions.items()}
            if o in function_lookup:
                return f'<PeekFunction {function_lookup[o]}>'
        return str(o)


def all_to_text(*args):
    fragments = []
    for v in args:
        fragments.extend(to_text(v))
    return ''.join([v[1] for v in fragments])


def to_text(val):
    # Normal lists which are not instances of `FormattedText` are
    # considered plain text.
    if isinstance(val, list) and not isinstance(val, FormattedText):
        return to_formatted_text(f"{val}")
    return to_formatted_text(val, auto_convert=True)



from typing import NamedTuple

from pygments.token import _TokenType

HTTP_METHODS = ['get', 'post', 'put', 'delete', 'head']
AUTO_SAVE_NAME = '__auto__'
DEFAULT_SAVE_NAME = '__default__'


class AlwaysNoneNameSpace:
    def __getattr__(self, name):
        return None


class PeekToken(NamedTuple):
    index: int
    ttype: _TokenType
    value: str


NONE_NS = AlwaysNoneNameSpace()



import json
import logging
import os

_logger = logging.getLogger(__name__)


def load_json_specs(kibana_dir):
    oss_path = os.path.join(kibana_dir, 'src', 'plugins', 'console', 'server', 'lib', 'spec_definitions')
    xpack_path = os.path.join(
        kibana_dir, 'x-pack', 'plugins', 'console_extensions', 'server', 'lib', 'spec_definitions'
    )
    specs = _do_load_json_specs(os.path.join(oss_path, 'json'))
    specs.update(_do_load_json_specs(os.path.join(xpack_path, 'json')))
    return specs


def _do_load_json_specs(base_dir):
    _logger.debug(f'Loading json specs from: {base_dir!r}')
    specs = {}
    for sub_dir in ('generated', 'overrides'):
        d = os.path.join(base_dir, sub_dir)
        if not os.path.exists(d):
            _logger.warning(f'JSON specs directory does not exist: {d}')
            continue
        for f in os.listdir(d):
            if f == '_common.json':
                continue
            with open(os.path.join(d, f)) as ins:
                spec = json.load(ins)
            if sub_dir == 'generated':
                specs.update(spec)
            else:
                for k, v in spec.items():
                    if k in specs:
                        specs[k].update(v)
                    else:
                        if k.startswith('xpack.'):
                            specs[k[6:]].update(v)
                        else:
                            specs['xpack.' + k].update(v)
    return specs



import ast
import itertools
import json
import logging
import operator
import os
import subprocess
import sys
import urllib
from numbers import Number
from subprocess import Popen

from elastic_transport._transport import TransportApiResponse
from pygments.token import Name

from peek.ast import (
    ArrayNode,
    BinOpNode,
    DictNode,
    EsApiCallFilePayloadNode,
    EsApiCallInlinePayloadNode,
    EsApiCallNode,
    ForInNode,
    FuncCallNode,
    GroupNode,
    KeyValueNode,
    LetNode,
    NameNode,
    Node,
    NumberNode,
    ShellOutNode,
    StringNode,
    SymbolNode,
    TextNode,
    UnaryOpNode,
    Visitor,
)
from peek.config import config_location
from peek.errors import PeekError
from peek.natives import EXPORTS
from peek.visitors import Ref

_logger = logging.getLogger(__name__)


def dot(left_operand, right_operand):
    if isinstance(left_operand, dict):
        if right_operand in left_operand:
            return left_operand[right_operand]
        else:
            raise PeekError(f'Value {left_operand!r} does not have key {right_operand!r}')
    elif isinstance(left_operand, list):
        if isinstance(right_operand, int):
            return left_operand[right_operand]
        else:
            raise PeekError(f'Cannot index array {left_operand!r} with non-integer value: {right_operand!r}')
    else:
        raise PeekError(f'Value {left_operand!r} must be either an array or dict')


def add(left_operand, right_operand):
    if isinstance(left_operand, str) and isinstance(right_operand, Number):
        return operator.add(left_operand, str(right_operand))
    elif isinstance(left_operand, Number) and isinstance(right_operand, str):
        return operator.add(str(left_operand), right_operand)
    else:
        return operator.add(left_operand, right_operand)


_BIN_OP_FUNCS = {
    '+': add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.truediv,
    '%': operator.mod,
    '.': dot,
}

_UNARY_OP_FUNCS = {
    '+': operator.pos,
    '-': operator.neg,
}


class PeekVM(Visitor):
    def __init__(self, app, bin_op_funcs=None, unary_op_funcs=None):
        super().__init__()
        self.app = app
        self._bin_op_funcs = bin_op_funcs or _BIN_OP_FUNCS
        self._unary_op_funcs = unary_op_funcs or _UNARY_OP_FUNCS
        self.context = {}
        self._load_context_file()
        self.builtins = EXPORTS
        if self.app.config.as_bool('load_extension'):
            self._load_extensions()

    @property
    def functions(self):
        return {k: v for k, v in itertools.chain(self.builtins.items(), self.context.items()) if callable(v)}

    def execute_node(self, node: Node):
        node.accept(self)

    def visit_es_api_call_node(self, node: EsApiCallNode):
        if isinstance(node.path_node, TextNode):
            path = node.path
        else:
            path_ref = Ref()
            with self.consumer(lambda v: path_ref.set(v)):
                node.path_node.accept(self)
            path = path_ref.get()
            path = path if path.startswith('/') else ('/' + path)

        options = Ref()
        with self.consumer(lambda v: options.set(v)):
            self._do_visit_dict_node(node.options_node)
        options = options.get()

        if isinstance(node, EsApiCallInlinePayloadNode):
            dicts = []
            with self.consumer(lambda v: dicts.append(v)):
                for dict_node in node.dict_nodes:
                    dict_node.accept(self)
            lines = [json.dumps(d) for d in dicts]
            payload = ('\n'.join(lines) + '\n') if lines else None
        elif isinstance(node, EsApiCallFilePayloadNode):
            f_ref = Ref()
            with self.consumer(lambda v: f_ref.set(v)):
                node.file_node.accept(self)
            with open(os.path.expanduser(f_ref.get().strip())) as ins:
                payload = ins.read()
                if self.app.config.as_bool('parse_payload_file'):
                    dicts = []
                    with self.consumer(lambda v: dicts.append(v)):
                        # NOTE this reuses the parser from the main app. It is not a problem
                        # because parser always finishes its job before returning. So in a
                        # single thread execution model, we won't corrupt the internals.
                        for pnode in self.app.parser.parse(payload, payload_only=True):
                            self.execute_node(pnode)
                    lines = [json.dumps(d) for d in dicts]
                    payload = ('\n'.join(lines) + '\n') if lines else None
                elif not payload.endswith('\n'):
                    payload += '\n'
        else:
            raise ValueError(f'Unknown node: {node!r}')

        conn = options.pop('conn') if 'conn' in options else None
        headers = options.pop('headers') if 'headers' in options else {}
        xoid = options.pop('xoid') if 'xoid' in options else None
        if xoid:
            headers['x-opaque-id'] = str(xoid)
        runas = options.pop('runas') if 'runas' in options else None
        if runas is not None:
            headers['es-security-runas-user'] = runas
        if conn is not None:
            es_client = self.app.es_client_manager.get_client(conn)
        else:
            es_client = self.app.es_client_manager.current
        outfile = options.pop('out', None)
        # Default to suppress on screen output if output file is provided
        quiet = options.pop('quiet', outfile is not None)
        pipe = options.pop('pipe', None)

        if options:
            self.app.display.error(f'Unknown options: {options}')
            return

        try:
            final_path = _maybe_encode_date_math(path)
            final_headers = headers if headers else None
            self.context['__'] = {
                'method': node.method,
                'path': final_path,
                'payload': payload,
                'headers': final_headers,
            }
            response: TransportApiResponse = es_client.perform_request(
                node.method, final_path, payload, headers=final_headers
            )

            warning = response.meta.headers.get('warning')
            if warning is not None and self.app.config.as_bool('show_warnings'):
                # warning header has the format of "299 buildInfo message"
                self.app.display.warn(warning.split(" ", 2)[2])

            out = response.body
            if pipe:
                process = Popen(pipe, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                process.stdin.write(response.body.encode('utf-8'))
                out, err = process.communicate()
                out = out.decode('utf-8')
                err = err.decode('utf-8')
                if err:
                    raise ValueError(f'{response.body}\n{err}')

            self.context['_'] = _maybe_decode_json(out)
            if outfile is not None:
                with open(outfile, 'w') as outs:
                    outs.write(out)
            if not quiet:
                self.app.display.info(out, header_text=self._get_header_text(response.meta, conn, runas))
        except Exception as e:
            if getattr(e, 'info', None) is not None and isinstance(getattr(e, 'status_code', None), int):
                self.context['_'] = _maybe_decode_json(e.info) if isinstance(e.info, str) else e.info
                self.app.display.info(e.info, header_text=self._get_header_text(conn, runas))
            else:
                self.app.display.error(getattr(e, 'message', str(e)), header_text=self._get_header_text(conn, runas))
                _logger.exception(f'Error on ES API call: {node!r}')

    def visit_func_call_node(self, node: FuncCallNode):
        if isinstance(node.name_node, NameNode):
            func = self.get_value(node.name_node.token.value)
        else:
            func_ref = Ref()
            with self.consumer(lambda v: func_ref.set(v)):
                node.name_node.accept(self)
            func = func_ref.get()
        if not callable(func):
            raise PeekError(f'{node.name_node!r} is not a callable, but {func!r}')

        func_symbols = Ref()
        with self.consumer(lambda v: func_symbols.set(v)):
            node.symbols_node.accept(self)

        func_args = Ref()
        with self.consumer(lambda v: func_args.set(v)):
            node.args_node.accept(self)

        for kv_node in node.kwargs_node.kv_nodes:
            assert isinstance(kv_node.key_node, NameNode), f'{kv_node.key_node!r}'
        func_kwargs = Ref()
        with self.consumer(lambda v: func_kwargs.set(v)):
            self._do_visit_dict_node(node.kwargs_node, resolve_key_name=False)
        kwargs = func_kwargs.get()
        if func_symbols.get():
            kwargs['@'] = func_symbols.get()
        try:
            result = func(self.app, *func_args.get(), **kwargs)
            if node.is_stmt:
                self.app.display.info(result)
            else:
                self.consume(result)
        except Exception as e:
            _logger.exception(f'Error on invoking function: {node.name_node!r}')
            self.app.display.error(e)

    def visit_let_node(self, node: LetNode):
        for kv_node in node.assignments_node.kv_nodes:
            lhs_chain = []
            with self.consumer(lambda v: lhs_chain.append(v)):
                self._unwind_lhs(kv_node.key_node)

            rhs = Ref()
            with self.consumer(lambda v: rhs.set(v)):
                kv_node.value_node.accept(self)

            if len(lhs_chain) == 1:
                self.context[lhs_chain[0]] = rhs.get()
            else:
                if lhs_chain[0] not in self.context:
                    raise PeekError(f'Unknown name: {lhs_chain[0]!r}')
                lhs = self.context[lhs_chain[0]]
                for x in lhs_chain[1:-1]:
                    if isinstance(lhs, dict) or (isinstance(lhs, list) and isinstance(x, int)):
                        lhs = lhs[x]
                    else:
                        raise PeekError(f'Invalid lhs for assignment: {lhs_chain}')

                x = lhs_chain[-1]
                if isinstance(lhs, dict) or (isinstance(lhs, list) and isinstance(x, int)):
                    lhs[x] = rhs.get()
                else:
                    raise PeekError(f'Invalid lhs for assignment: {lhs_chain}')

    def visit_shell_out_node(self, node: ShellOutNode):
        try:
            input_fd = self.app.prompt.input.fileno()
        except AttributeError:
            input_fd = sys.stdin.fileno()
        try:
            output_fd = self.app.prompt.output.fileno()
        except AttributeError:
            output_fd = sys.stdout.fileno()
        p = Popen(node.command, shell=True, stdin=input_fd, stdout=output_fd)
        p.wait()

    def visit_for_in_node(self, node: ForInNode):
        var_name = node.item.token.value
        items_ref = Ref()
        with self.consumer(lambda v: items_ref.set(v)):
            node.items.accept(self)
        items = items_ref.get()
        if not isinstance(items, list):
            raise PeekError(f'For in loop must operator over a list, got {items!r}')

        for i in items:
            self.context[var_name] = i
            for n in node.suite:
                n.accept(self)

    def visit_key_value_node(self, node: KeyValueNode):
        node.key_node.accept(self)
        node.value_node.accept(self)

    def visit_name_node(self, node: NameNode):
        v = self.get_value(node.token.value)
        self.consume(v)

    def visit_symbol_node(self, node: SymbolNode):
        self.consume(node.token.value)

    def visit_string_node(self, node: StringNode):
        self.consume(ast.literal_eval(node.token.value))

    def visit_number_node(self, node: NumberNode):
        self.consume(ast.literal_eval(node.token.value))

    def visit_dict_node(self, node: DictNode):
        self._do_visit_dict_node(node, resolve_key_name=True)

    def visit_array_node(self, node: ArrayNode):
        values = []
        with self.consumer(lambda v: values.append(v)):
            for node in node.value_nodes:
                node.accept(self)
        self.consume(values)

    def visit_text_node(self, node: TextNode):
        if node.token.ttype is Name.Builtin:
            self.consume({'true': True, 'false': False, 'null': None}[node.token.value])
        else:
            self.consume(node.token.value)

    def visit_bin_op_node(self, node: BinOpNode):
        left_operand = Ref()
        with self.consumer(lambda v: left_operand.set(v)):
            node.left_node.accept(self)

        right_operand = Ref()
        with self.consumer(lambda v: right_operand.set(v)):
            node.right_node.accept(self)

        op_func = self._bin_op_funcs.get(node.op_token.value, None)
        if op_func is None:
            raise PeekError(f'Unknown binary operator: {node.op_token.value!r}')
        self.consume(op_func(left_operand.get(), right_operand.get()))

    def visit_unary_op_node(self, node: UnaryOpNode):
        operand = Ref()
        with self.consumer(lambda v: operand.set(v)):
            node.operand_node.accept(self)

        op_func = self._unary_op_funcs.get(node.op_token.value, None)
        if op_func is None:
            raise PeekError(f'Unknown unary operator: {node.op_token.value!r}')
        self.consume(op_func(operand.get()))

    def visit_group_node(self, node: GroupNode):
        node.grouped.accept(self)

    def _do_visit_dict_node(self, node: DictNode, resolve_key_name=False):
        keys = []
        values = []
        with self.consumer(lambda v: keys.append(v)):
            for kv_node in node.kv_nodes:
                if resolve_key_name or not isinstance(kv_node.key_node, NameNode):
                    kv_node.key_node.accept(self)
                else:
                    self.consume(kv_node.key_node.token.value)
                with self.consumer(lambda v: values.append(v)):
                    kv_node.value_node.accept(self)
        assert len(keys) == len(values), f'key length is not equal to value length: {keys!r}, {values!r}'
        self.consume(dict(zip(keys, values)))

    def get_value(self, name):
        value = self.builtins.get(name)
        if value is None:
            value = self.context.get(name)
        if value is None:
            raise NameError(f'Unknown name: {name!r}')
        return value

    def _unwind_lhs(self, node: Node):
        if isinstance(node, NameNode):
            self.consume(node.token.value)
        elif isinstance(node, BinOpNode):
            if node.op_token.value != '.':
                raise PeekError(f'lhs can only have variable and dot notation, but got {node!r}')
            self._unwind_lhs(node.left_node)
            node.right_node.accept(self)
        else:
            raise PeekError(f'lhs can only have variable and dot notation, but got {node!r}')

    def _load_context_file(self):
        context_file = self.app.config['context_file']
        if not context_file:
            return

        ns = {}
        with open(context_file) as ins:
            exec(ins.read(), {}, ns)
        self.context.update(ns)

    def _load_extensions(self):
        """
        Load extra variables from external paths
        """
        default_extension_path = os.path.join(config_location(), 'extensions')
        if os.path.exists(default_extension_path):
            self._load_one_extension_path(default_extension_path)

        extension_path = self.app.config['extension_path']
        if not extension_path:
            return

        sys_path = sys.path[:]
        try:
            for p in extension_path.split(os.pathsep):
                if os.path.isfile(p):
                    self._load_one_extension_file(p)
                elif os.path.isdir(p):
                    self._load_one_extension_path(p)
                else:
                    _logger.warning(f'Cannot load extension path: {p}')
        finally:
            sys.path = sys_path

    def _load_one_extension_path(self, p):
        for f in os.listdir(p):
            if not f.endswith('.py'):
                continue
            self._load_one_extension_file(os.path.join(p, f))

    def _load_one_extension_file(self, p):
        _logger.info(f'Loading extension: {p!r}')
        import importlib

        fields = os.path.splitext(p)
        if len(fields) != 2 or fields[1] != '.py':
            _logger.warning(f'Extension must be python files, got: {p!r}')
            return
        sys.path.insert(0, os.path.dirname(fields[0]))
        try:
            module_name = os.path.basename(fields[0])
            if module_name in sys.modules:
                m = importlib.reload(sys.modules[module_name])
            else:
                m = importlib.import_module(module_name)
            exports = getattr(m, 'EXPORTS', None)
            if isinstance(exports, dict):
                self.context.update(exports)
                _logger.info(f'Extension loaded: {p!r}')
            else:
                _logger.warning(f'Ignore extension {p!r} since EXPORTS is not a dict, but: {exports!r}')
        except Exception as e:
            _logger.error(f'Error on loading extension: {p!r}, {e}')
            _logger.exception(f'Error on loading extension: {p!r}')

    def _get_header_text(self, meta, conn, runas):
        parts = []
        if conn is not None:
            parts.append(f'conn={conn!r}')
        if runas is not None:
            parts.append(f'runas={runas!r}')

        parts.append(f'took={meta.duration * 1000:.3f}ms')
        return ' '.join(parts)


def _maybe_decode_json(r):
    try:
        return json.loads(r)
    except Exception:
        return r


def _maybe_encode_date_math(path):
    parts = []
    current_pos = 0
    while True:
        next_pos = path.find('}>', current_pos)
        if next_pos == -1:
            parts.append(path[current_pos:])
            break
        next_pos += 2
        pos = path.rfind('<', current_pos, next_pos)
        if pos > current_pos:
            parts.append(path[current_pos:pos])
        parts.append(urllib.parse.quote(path[pos:next_pos], safe=''))
        current_pos = next_pos

    return ''.join(parts)



import ast
import json
import logging
from enum import Enum
from typing import Callable, Iterable, NamedTuple

from pygments.token import Comment, Error, Literal, Name, Number, String, Token, Whitespace

from peek.ast import (
    ArrayNode,
    BinOpNode,
    DictNode,
    EsApiCallFilePayloadNode,
    EsApiCallInlinePayloadNode,
    ForInNode,
    FuncCallNode,
    GroupNode,
    KeyValueNode,
    LetNode,
    NameNode,
    NumberNode,
    ShellOutNode,
    StringNode,
    SymbolNode,
    TextNode,
    UnaryOpNode,
)
from peek.common import HTTP_METHODS, PeekToken
from peek.errors import PeekSyntaxError
from peek.lexers import (
    EOF,
    Assign,
    At,
    BinOp,
    BlankLine,
    BracketLeft,
    BracketRight,
    Colon,
    Comma,
    CurlyLeft,
    CurlyRight,
    DictKey,
    For,
    FuncName,
    HttpMethod,
    In,
    Let,
    OptionName,
    ParenLeft,
    ParenRight,
    PeekLexer,
    ShellOut,
    TripleD,
    TripleS,
    UnaryOp,
)

_logger = logging.getLogger(__name__)

HTTP_METHODS = [m.upper() for m in HTTP_METHODS]

_BIN_OP_ORDERS = {
    None: -1,
    '+': 100,
    '-': 100,
    '*': 200,
    '/': 200,
    '%': 200,
    '.': 300,
}


class ParserEventType(Enum):
    ES_METHOD = 'ES_METHOD'
    ES_URL = 'ES_URL'
    BEFORE_ES_URL_EXPR = 'BEFORE_ES_URL_EXPR'
    AFTER_ES_URL_EXPR = 'AFTER_ES_URL_EXPR'
    ES_OPTION_NAME = 'ES_OPTION_NAME'
    BEFORE_ES_OPTION_VALUE = 'BEFORE_ES_OPTION_VALUE'
    AFTER_ES_OPTION_VALUE = 'AFTER_ES_OPTION_VALUE'
    ES_PAYLOAD_FILE_AT = 'ES_PAYLOAD_FILE_AT'
    BEFORE_ES_PAYLOAD_INLINE = 'BEFORE_ES_PAYLOAD_INLINE'
    AFTER_ES_PAYLOAD_INLINE = 'AFTER_ES_PAYLOAD_INLINE'
    FUNC_STMT = 'FUNC_STMT'
    BEFORE_FUNC_ARGS = 'BEFORE_FUNC_ARGS'
    AFTER_FUNC_ARGS = 'AFTER_FUNC_ARGS'
    BEFORE_FUNC_REGULAR_ARG = 'BEFORE_FUNC_REGULAR_ARG'
    AFTER_FUNC_REGULAR_ARG = 'AFTER_FUNC_REGULAR_ARG'
    BEFORE_FUNC_SYMBOL_ARG = 'BEFORE_FUNC_SYMBOL_ARG'
    AFTER_FUNC_SYMBOL_ARG = 'AFTER_FUNC_SYMBOL_ARG'
    FUNC_OPTION_NAME_OR_ARG = 'FUNC_OPTION_NAME_OR_ARG'
    BEFORE_FUNC_OPTION_VALUE = 'BEFORE_FUNC_OPTION_VALUE'
    AFTER_FUNC_OPTION_VALUE = 'AFTER_FUNC_OPTION_VALUE'
    SHELL_OUT = 'SHELL_OUT'
    LET = 'LET'
    FOR = 'FOR'
    BEFORE_FOR_SUITE = 'BEFORE_FOR_SUITE'
    AFTER_FOR_SUITE = 'AFTER_FOR_SUITE'
    DICT_KEY = 'DICT_KEY'
    BEFORE_DICT_KEY_EXPR = 'BEFORE_DICT_KEY_EXPR'
    AFTER_DICT_KEY_EXPR = 'AFTER_DICT_KEY_EXPR'
    BEFORE_DICT_VALUE = 'BEFORE_DICT_VALUE'
    AFTER_DICT_VALUE = 'AFTER_DICT_VALUE'
    AFTER_TOKEN = 'AFTER_TOKEN'


# Event to be published right before the actual parsing
class ParserEvent(NamedTuple):
    type: ParserEventType
    token: PeekToken


class PeekParser:
    """
    The parser is not thread safe and does not pretend to be.
    """

    def __init__(self, listeners=None):
        self.lexer = PeekLexer()
        self.text = ''
        self.position = 0  # position is token position, not character
        self.tokens = []
        self.listeners: Iterable[Callable] = listeners or []

    def parse(self, text, payload_only=False, fail_fast_on_error_token=True, last_stmt_only=False, log_level=None):
        saved_log_level = _logger.getEffectiveLevel()
        try:
            if log_level is not None:
                _logger.setLevel(log_level)
            self.text = text
            self.position = 0
            self.tokens = []

            stack = ('dict',) if payload_only else ('root',)
            self.tokens = process_tokens(self.lexer.get_tokens_unprocessed(self.text, stack=stack))
            if last_stmt_only:
                idx_last_stmt_token = find_last_stmt_token(self.tokens)
                if idx_last_stmt_token == -1:
                    return []
                else:
                    self.position = idx_last_stmt_token

            if fail_fast_on_error_token:
                for token in self.tokens[self.position :]:
                    if token.ttype in Token.Error:
                        raise PeekSyntaxError(self.text, token, message='Found error token and fail fast is enabled')

            return self._do_parse_payload() if payload_only else self._do_parse()
        finally:
            if log_level is not None:
                _logger.setLevel(saved_log_level)

    def _do_parse(self):
        nodes = []
        while self._peek_token().ttype is not EOF:
            token = self._peek_token()
            if token.ttype is BlankLine:
                self._consume_token(BlankLine)
            else:
                nodes.append(self._parse_stmt())
        return nodes

    def _do_parse_payload(self):
        nodes = []
        while self._peek_token().ttype is not EOF:
            nodes.append(self._parse_dict())
        return nodes

    def _parse_stmt(self):
        token = self._peek_token()
        if token.ttype is FuncName:
            return self._parse_func_call()
        elif token.ttype is HttpMethod:
            return self._parse_es_api_call()
        elif token.ttype is Let:
            return self._parse_let_stmt()
        elif token.ttype is ShellOut:
            return self._parse_shell_out()
        elif token.ttype is For:
            return self._parse_for_stmt()
        else:
            raise PeekSyntaxError(self.text, token, title='Invalid token', message='Expect beginning of a statement')

    def _parse_es_api_call(self):
        self._publish_event(ParserEventType.ES_METHOD)
        method_token = self._consume_token(HttpMethod)
        method_node = NameNode(method_token)
        if method_token.value.upper() not in HTTP_METHODS:
            raise PeekSyntaxError(
                self.text,
                method_token,
                title='Invalid HTTP method',
                message=f'Expect HTTP method of value in {HTTP_METHODS!r}, got {method_token.value!r}',
            )

        if self._peek_token().ttype is Literal:
            self._publish_event(ParserEventType.ES_URL)
            path_node = TextNode(self._consume_token(Literal))
        elif self._peek_token().ttype is ParenLeft:
            self._publish_event(ParserEventType.BEFORE_ES_URL_EXPR)
            path_node = self._parse_expr()
            self._publish_event(ParserEventType.AFTER_ES_URL_EXPR)
        else:
            raise PeekSyntaxError(
                self.text,
                self._peek_token(),
                message='HTTP path must be either text literal or an expression enclosed by parenthesis',
            )

        option_nodes = []
        while self._peek_token().ttype is OptionName:
            self._publish_event(ParserEventType.ES_OPTION_NAME)
            n = NameNode(self._consume_token(OptionName))
            self._consume_token(Assign)
            self._publish_event(ParserEventType.BEFORE_ES_OPTION_VALUE)
            option_nodes.append(KeyValueNode(n, self._parse_expr()))
            self._publish_event(ParserEventType.AFTER_ES_OPTION_VALUE)

        if self._peek_token().ttype is At:
            self._publish_event(ParserEventType.ES_PAYLOAD_FILE_AT)
            self._consume_token(At)
            return EsApiCallFilePayloadNode(
                method_node, path_node, DictNode(option_nodes), SymbolNode(self._consume_token(Literal))
            )
        else:
            dict_nodes = []
            while self._peek_token().ttype is not EOF:
                if self._peek_token().ttype is CurlyLeft:
                    self._publish_event(ParserEventType.BEFORE_ES_PAYLOAD_INLINE)
                    dict_nodes.append(self._parse_dict())
                    self._publish_event(ParserEventType.AFTER_ES_PAYLOAD_INLINE)
                else:
                    break
            return EsApiCallInlinePayloadNode(method_node, path_node, DictNode(option_nodes), dict_nodes)

    def _parse_let_stmt(self):
        self._publish_event(ParserEventType.LET)
        self._consume_token(Let)
        kv_nodes = []
        while self._peek_token().ttype is not EOF:
            if self._peek_token().ttype is BlankLine:
                self._consume_token(BlankLine)
                break
            key_node = self._parse_expr()
            self._consume_token(Assign)
            value_node = self._parse_expr()
            kv_nodes.append(KeyValueNode(key_node, value_node))
        return LetNode(DictNode(kv_nodes))

    def _parse_for_stmt(self):
        self._publish_event(ParserEventType.FOR)
        self._consume_token(For)
        item = NameNode(self._consume_token(Name))
        self._consume_token(In)
        items = self._parse_expr()
        self._publish_event(ParserEventType.BEFORE_FOR_SUITE)
        self._consume_token(CurlyLeft)
        suite = []
        while self._peek_token().ttype is not CurlyRight:
            token = self._peek_token()
            if token.ttype is BlankLine:
                self._consume_token(BlankLine)
            else:
                suite.append(self._parse_stmt())
        self._consume_token(CurlyRight)
        self._publish_event(ParserEventType.AFTER_FOR_SUITE)
        return ForInNode(item, items, suite)

    def _parse_dict(self):
        kv_nodes = []
        self._consume_token(CurlyLeft)
        while self._peek_token().ttype is not CurlyRight:
            kv_nodes.append(self._parse_dict_kv())
            if self._peek_token().ttype is Comma:
                self._consume_token(Comma)
            else:
                break
        self._consume_token(CurlyRight)
        return DictNode(kv_nodes)

    def _parse_dict_kv(self):
        if self._peek_token().ttype is DictKey:
            self._publish_event(ParserEventType.DICT_KEY)
            key_node = StringNode(self._consume_token(DictKey))
        else:
            self._publish_event(ParserEventType.BEFORE_DICT_KEY_EXPR)
            key_node = self._parse_expr()
            self._publish_event(ParserEventType.AFTER_DICT_KEY_EXPR)
        self._consume_token(Colon)
        self._publish_event(ParserEventType.BEFORE_DICT_VALUE)
        value_node = self._parse_expr()
        self._publish_event(ParserEventType.AFTER_DICT_VALUE)
        return KeyValueNode(key_node, value_node)

    def _parse_array(self):
        value_nodes = []
        self._consume_token(BracketLeft)
        while self._peek_token().ttype is not BracketRight:
            value_nodes.append(self._parse_expr())
            if self._peek_token().ttype is Comma:
                self._consume_token(Comma)
            else:
                break
        self._consume_token(BracketRight)
        return ArrayNode(value_nodes)

    def _parse_value(self):
        token = self._peek_token()
        if token.ttype is CurlyLeft:
            return self._parse_dict()
        elif token.ttype is BracketLeft:
            return self._parse_array()
        elif token.ttype in Number:
            return NumberNode(self._consume_token(token.ttype))
        elif token.ttype is Name.Builtin:
            return TextNode(self._consume_token(token.ttype))
        elif token.ttype in (String.Double, String.Single, TripleS, TripleD):
            return StringNode(self._consume_token(token.ttype))
        else:
            raise PeekSyntaxError(self.text, token, message=f'Unexpected token when parsing for value: {token!r}')

    def _parse_func_call(self):
        self._publish_event(ParserEventType.FUNC_STMT)
        name_node = NameNode(self._consume_token(FuncName))
        symbol_nodes, arg_nodes, kwarg_nodes = self._parse_func_call_args()
        return FuncCallNode(name_node, ArrayNode(symbol_nodes), ArrayNode(arg_nodes), DictNode(kwarg_nodes))

    def _parse_func_call_args(self, is_stmt=True):
        self._publish_event(ParserEventType.BEFORE_FUNC_ARGS)
        symbol_nodes = []
        arg_nodes = []
        kwarg_nodes = []
        while self._peek_token().ttype is not EOF:
            if self._peek_token().ttype is BlankLine:
                self._consume_token(BlankLine)
                if is_stmt:
                    break
            elif self._peek_token().ttype is ParenRight:
                if is_stmt:
                    raise PeekSyntaxError(
                        self.text,
                        self._peek_token(),
                        message='Found function expression while parsing for function stmt',
                    )
                else:
                    break
            elif self._peek_token().ttype is Name:
                self._publish_event(ParserEventType.FUNC_OPTION_NAME_OR_ARG)
                n = NameNode(self._consume_token(Name))
                if self._peek_token().ttype is Assign:
                    self._consume_token(Assign)
                    self._publish_event(ParserEventType.BEFORE_FUNC_OPTION_VALUE)
                    kwarg_nodes.append(KeyValueNode(n, self._parse_expr()))
                    self._publish_event(ParserEventType.AFTER_FUNC_OPTION_VALUE)
                else:
                    arg_nodes.append(self._parse_expr_after_left_operand(n))
            elif self._peek_token().ttype is At:
                self._publish_event(ParserEventType.BEFORE_FUNC_SYMBOL_ARG)
                self._consume_token(At)
                symbol_nodes.append(SymbolNode(self._consume_token(Literal)))
                self._publish_event(ParserEventType.AFTER_FUNC_SYMBOL_ARG)
            else:
                self._publish_event(ParserEventType.BEFORE_FUNC_REGULAR_ARG)
                arg_nodes.append(self._parse_expr())
                self._publish_event(ParserEventType.AFTER_FUNC_REGULAR_ARG)
        self._publish_event(ParserEventType.AFTER_FUNC_ARGS)
        return symbol_nodes, arg_nodes, kwarg_nodes

    def _parse_shell_out(self):
        self._publish_event(ParserEventType.SHELL_OUT)
        self._consume_token(ShellOut)
        return ShellOutNode(TextNode(self._consume_token(Literal)))

    def _parse_expr(self, last_bin_op=None):
        if self._peek_token().ttype is UnaryOp:
            unary_op_token = self._consume_token(UnaryOp)
        else:
            unary_op_token = None

        if self._peek_token().ttype is ParenLeft:
            pl = self._consume_token(ParenLeft)
            n = self._parse_expr()
            pr = self._consume_token(ParenRight)
            n = GroupNode(n, pl, pr)
        elif self._peek_token().ttype is Name:
            n = NameNode(self._consume_token(Name))
        elif self._peek_token().ttype is At:
            self._consume_token(At)
            n = SymbolNode(self._consume_token(Literal))
        else:
            n = self._parse_value()

        return self._parse_expr_after_left_operand(n, unary_op_token=unary_op_token, last_bin_op=last_bin_op)

    def _parse_expr_after_left_operand(self, n, unary_op_token=None, last_bin_op=None):
        while True:
            if self._peek_token().ttype is BinOp:
                op_token = self._peek_token()
                if _BIN_OP_ORDERS[last_bin_op] >= _BIN_OP_ORDERS[op_token.value]:
                    return n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
                else:
                    self._consume_token(op_token.ttype)
                    right_node = self._parse_expr(last_bin_op=op_token.value)
                    n = n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
                    n = BinOpNode(op_token, n, right_node)
            elif self._peek_token().ttype is ParenLeft:  # func call
                if last_bin_op == '.':
                    return n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
                else:
                    self._consume_token(ParenLeft)
                    symbol_nodes, arg_nodes, kwarg_nodes = self._parse_func_call_args(is_stmt=False)
                    self._consume_token(ParenRight)
                    n = FuncCallNode(
                        n, ArrayNode(symbol_nodes), ArrayNode(arg_nodes), DictNode(kwarg_nodes), is_stmt=False
                    )
                    n = n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
            else:
                return n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
            unary_op_token = None

    def _peek_token(self) -> PeekToken:
        if self.position >= len(self.tokens):
            return PeekToken(len(self.text), EOF, '\0')
        return self.tokens[self.position]

    def _consume_token(self, ttype, value=None) -> PeekToken:
        token = self._peek_token()
        self.position += 1
        if token.ttype is not ttype:
            raise PeekSyntaxError(self.text, token, message=f'Expect token of type {ttype!r}, got {token.ttype!r}')
        if value and (token.value != value or token.value not in value):
            raise PeekSyntaxError(self.text, token, message=f'Expect token of value {value!r}, got {token.value!r}')
        self._publish_event(ParserEventType.AFTER_TOKEN, token)
        return token

    def _publish_event(self, event_type: ParserEventType, token=None):
        token = token or self._peek_token()
        for listener in self.listeners:
            listener(ParserEvent(event_type, token))


def normalise_string(value):
    return json.dumps(ast.literal_eval(value))


def process_tokens(tokens: Iterable[PeekToken]):
    """
    Process tokens by filtering out whitespaces and merging tokens that
    should be represented as one, e.g. Strings, BlankLine, Error.
    """
    processed_tokens = []
    current_token = None
    for token in tokens:
        if token.ttype in (Whitespace, Comment.Single):
            if current_token is not None:
                processed_tokens.append(current_token)
                current_token = None
        elif token.ttype in String or token.ttype is BlankLine or token.ttype is Error:
            if current_token is not None and current_token.ttype is token.ttype:
                current_token = PeekToken(current_token.index, current_token.ttype, current_token.value + token.value)
            else:
                # two consecutive strings with different quotes should not be merged
                if current_token is not None:
                    processed_tokens.append(current_token)
                current_token = token
        else:
            if current_token is not None:
                processed_tokens.append(current_token)
                current_token = None
            processed_tokens.append(token)
    if current_token is not None:
        processed_tokens.append(current_token)

    return processed_tokens


def find_last_stmt_token(tokens) -> int:
    """
    Find the last token that can start a statement
    """
    for i in range(len(tokens) - 1, -1, -1):
        if tokens[i].ttype in (HttpMethod, FuncName, ShellOut, Let, For):
            return i
    return -1
